# -*- coding: utf-8 -*-

#Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
A collection of utilities for retrieving parameters that may be relevant 
for ionospheric corrections.
"""

import os
import glob
import gzip
import numpy
import socket
import shutil
import logging
import tarfile
import tempfile
import subprocess
try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen
from datetime import datetime, timedelta

from scipy.special import lpmv
from scipy.misc import factorial
from scipy.optimize import fmin
from scipy.interpolate import RectBivariateSpline

from lsl.common.stations import geo2ecef
from lsl.common.paths import data as dataPath
from lsl.common.mcs import mjdmpm2datetime, datetime2mjdmpm


__version__ = "0.5"
__revision__ = "$Rev: 2620 $"
__all__ = ['getMagneticField', 'computeMagneticDeclination', 'computeMagneticInclination', 
        'getTECValue', 'getIonosphericPiercePoint', 
        '__version__', '__revision__', '__all__']


# Logger for capturing problems with downloading EOP data
__logger = logging.getLogger('__main__')


# Create the cache directory
if not os.path.exists(os.path.join(os.path.expanduser('~'), '.lsl')):
    os.mkdir(os.path.join(os.path.expanduser('~'), '.lsl'))
_CacheDir = os.path.join(os.path.expanduser('~'), '.lsl', 'ionospheric_cache')
if not os.path.exists(_CacheDir):
    os.mkdir(_CacheDir)


# Create the on-line cache
_Cache = {}

# Radius of the Earth in meters for the IGRF
_rEarth = 6371.2e3


def _loadIGRFModel(filename):
    """
    Given a filename pointing to a list of IGRF coefficients, load in the 
    data and return a dictionary containing the raw coefficients.
    
    The dictionary keys are:
    * years - list of years for each of the models
    * g - dictionary of cosine term coefficients
    * h - dictionary of sine term coefficients
    
    The g and h dictionaries are keyed off the degree of the Legendre 
    function and each stores a list of n orders.  Each order is composed
    of len(years)+1 values, one for each each plus a secular evolution
    term.
    """
    
    # Open the file
    with open(filename, 'r') as fh:
        # Go!
        dataCos = {}
        dataSin = {}
        for line in fh:
            ## Is this line a comment?
            line = line.replace('\n', '')
            if line.find('#') != -1:
                continue
                
            ## Is it a header?
            if line.find('IGRF') != -1:
                continue
            if line.find('g/h') != -1:
                fields = line.split(None)
                years = [float(value) for value in fields[3:-1]]
                continue
                
            ## Must be data...  parse it
            fields = line.split(None)
            t, n, m = fields[0], int(fields[1]), int(fields[2])
            c = numpy.array([float(v) for v in fields[3:]])
            
            ## Sort out cosine (g) vs. sine (h)
            if t == 'g':
                try:
                    dataCos[n][m] = c
                except KeyError:
                    dataCos[n] = [numpy.zeros(len(years)+1) for i in xrange(n+1)]
                    dataCos[n][m] = c
            else:
                try:
                    dataSin[n][m] = c
                except KeyError:
                    dataSin[n] = [numpy.zeros(len(years)+1) for i in xrange(n+1)]
                    dataSin[n][m] = c
                    
    # Build the output
    output = {'years': years, 'g': dataCos, 'h': dataSin}
    
    # Done
    return output


def _computeIGRFCoefficents(year, coeffs):
    """
    Given a decimal year and a coefficient dictionary from _parseIGRFModel(),
    compute the actual coefficients for that epoch and return a dictionary
    containing the cosine and sine coefficients for that year.
    
    The dictionary keys are:
    * year - year the coefficients are computed for
    * g - dictionary of cosine term coefficients
    * h - dictionary of sine term coefficients
    
    The g and h dictionaries are keyed off the degree of the Legendre 
    function and each stores a list of n orders.
    """
    
    # Figure out the closest model point(s) to the requested year taking into
    # account that a new model comes out every five years
    best = numpy.where( numpy.abs(year - numpy.array(coeffs['years'])) < 5 )[0]
    
    if year < min(coeffs['years']):
        # If the requested year is before 1900 we can't do anything
        raise RuntimeError("Invalid year %i" % year)
    else:
        # Otherwise, figure out the coefficients using a simple linear interpolation 
        # or extrapolation using the secular changes in the model
        coeffsCos = {}
        coeffsSin = {}
        
        # Loop over the degrees
        for n in coeffs['g'].keys():
            ## Loop over the orders
            for m in xrange(0, n+1):
                if year > max(coeffs['years']):
                    ### If we are beyond the last year in the model, use the secular changes
                    slope = coeffs['g'][n][m][-1]
                else:
                    ### Otherwise, do a linear interpolation between the two nearest models
                    slope = coeffs['g'][n][m][best[-1]] - coeffs['g'][n][m][best[0]]
                    if best[-1] == best[0]:
                        slope = 0.0
                    else:
                        slope /= coeffs['years'][best[-1]] - coeffs['years'][best[0]]
                        
                ### Compute the cosine terms
                try:
                    coeffsCos[n][m] = slope*(year - coeffs['years'][best[0]]) + coeffs['g'][n][m][best[0]]
                except:
                    coeffsCos[n] = [0.0 for i in xrange(n+1)]
                    coeffsCos[n][m] = slope*(year - coeffs['years'][best[0]]) + coeffs['g'][n][m][best[0]]
                    
                if year > max(coeffs['years']):
                    ### If we are beyond the last year in the model, use the secular changes
                    slope = coeffs['h'][n][m][-1]
                else:
                    ### Otherwise, do a linear interpolation between the two nearest models
                    slope = coeffs['h'][n][m][best[-1]] - coeffs['h'][n][m][best[0]]
                    if best[-1] == best[0]:
                        slope = 0.0
                    else:
                        slope /= coeffs['years'][best[-1]] - coeffs['years'][best[0]]
                        
                ### Compute the sine terms
                try:
                    coeffsSin[n][m] = slope*(year - coeffs['years'][best[0]]) + coeffs['h'][n][m][best[0]]
                except:
                    coeffsSin[n] = [0.0 for i in xrange(n+1)]
                    coeffsSin[n][m] = slope*(year - coeffs['years'][best[0]]) + coeffs['h'][n][m][best[0]]
                    
    # Build the output
    output = {'year': year, 'g': coeffsCos, 'h': coeffsSin}
    
    # Done
    return output


def _Snm(n, m):
    """
    Compute the factor needed to convert an unnormalized associated Legendre
    function of degree n and order m into a Schmidt quasi-normalized
    associated Legendre function.
    """
    
    if m == 0:
        return numpy.sqrt(factorial(n-m)/factorial(n+m))
    else:
        return numpy.sqrt(2.0*factorial(n-m)/factorial(n+m))


def _Pnm(n, m, mu):
    """
    Compute the value of an unnormalized associated Legendre function of
    degree n and order m at mu following the convention of Abramowitz 
    and Stegun (1972).
    """
    
    return (-1)**m*lpmv(m, n, mu)



def _dPnm(n, m, mu):
    """
    Compute d Pnm(cos(theta)) / d theta for an unnormalized associated 
    Legendre function of degree n and order m at a cos(theta) of mu.
    """
    
    o = n*mu*_Pnm(n, m, mu) - (n+m)*_Pnm(n-1, m, mu)
    o /= numpy.sqrt(1.0 - mu**2)
    return o


def getMagneticField(lat, lng, elev, mjd=None, ecef=False):
    """
    Given a geodetic location described by a latitude in degrees (North 
    positive), a longitude in degrees (West negative), an elevation 
    in meters and an MJD value, compute the Earth's magnetic field in that 
    location and return a three-element tuple of the magnetic field's 
    components in nT.  By default these are in topocentric coordinates of
    (North, East, Up).  To return values in ECEF, set the 'ecef' keyword to
    True.  If the MJD file is None, the current time is used.
    
    .. note::
        The convention used for the topocentric coordinates differs
        from what the IGRF uses in the sense that the zenith direction
        points up rather than down.
    """
    
    # Get the current time if mjd is None
    if mjd is None:
        mjd, mpm = datetime2mjdmpm( datetime.utcnow() )
        mjd = mjd + mpm/1000.0/3600.0/24.0
        
    # Convert the MJD to a decimal year.  This is a bit tricky
    ## Break the MJD into an integer MJD and an MPM in order to build a datetime instance
    mpm = int((mjd - int(mjd))*24.0*3600.0*1000.0)
    mjd0 = mjdmpm2datetime(int(mjd), mpm)
    ## Convert the datetime instance to January 1
    mjd0 = mjd0.replace(month=1, day=1, hour=0, second=0, microsecond=0)
    ## Figure out January 1 for the following year
    mjd1 = mjd0.replace(year=mjd0.year+1)
    ## Figure out how long the year is in days
    diffDays = mjd1-mjd0
    diffDays = diffDays.days + diffDays.seconds/86400.0 + diffDays.microseconds/1e6/86400.0
    ## Convert the January 1 date back to an MJD
    mjd0, mpm0 = datetime2mjdmpm(mjd0)
    mjd0 = mjd0 + mpm/1000.0/3600.0/24.0
    year = (mjd1.year - 1) + (mjd - mjd0) / diffDays
    
    # Convert the geodetic position provided to a geocentric one for calculation
    ## Deal with the poles
    if 90.0 - lat < 0.001:
        xyz = numpy.array(geo2ecef(89.999*numpy.pi/180, lng*numpy.pi/180, elev))
    elif 90.0 + lat < 0.001:
        xyz = numpy.array(geo2ecef(-89.999*numpy.pi/180, lng*numpy.pi/180, elev))
    else:
        xyz = numpy.array(geo2ecef(lat*numpy.pi/180, lng*numpy.pi/180, elev))
    ## To geocentric
    r = numpy.sqrt( (xyz**2).sum() )
    lt = numpy.arcsin(xyz[2]/r)
    ln = numpy.arctan2(xyz[1], xyz[0])
    
    # Load in the coefficients
    try:
        coeffs = _Cache['IGRF']
    except KeyError:
        filename = os.path.join(dataPath, 'igrf12coeffs.txt')
        _Cache['IGRF'] = _loadIGRFModel(filename)
        
        coeffs = _Cache['IGRF']
        
    # Compute the coefficients for the epoch
    coeffs = _computeIGRFCoefficents(year, coeffs)
    
    # Compute the field strength in spherical coordinates
    Br, Bth, Bph = 0.0, 0.0, 0.0
    for n in coeffs['g'].keys():
        for m in xrange(0, n+1):
            Br  += (n+1.0)*(_rEarth/r)**(n+2) * _Snm(n,m)*coeffs['g'][n][m]*numpy.cos(m*ln) * _Pnm(n, m, numpy.sin(lt))
            Br  += (n+1.0)*(_rEarth/r)**(n+2) * _Snm(n,m)*coeffs['h'][n][m]*numpy.sin(m*ln) * _Pnm(n, m, numpy.sin(lt))
            
            Bth -= (_rEarth/r)**(n+2) * _Snm(n,m)*coeffs['g'][n][m]*numpy.cos(m*ln) * _dPnm(n, m, numpy.sin(lt))
            Bth -= (_rEarth/r)**(n+2) * _Snm(n,m)*coeffs['h'][n][m]*numpy.sin(m*ln) * _dPnm(n, m, numpy.sin(lt))
            
            Bph += (_rEarth/r)**(n+2)/numpy.cos(lt) * _Snm(n,m)*coeffs['g'][n][m]*m*numpy.sin(m*ln) * _Pnm(n, m, numpy.sin(lt))
            Bph -= (_rEarth/r)**(n+2)/numpy.cos(lt) * _Snm(n,m)*coeffs['h'][n][m]*m*numpy.cos(m*ln) * _Pnm(n, m, numpy.sin(lt))
    ## And deal with NaNs
    if Br != Br:
        Br = 0.0
    if Bth != Bth:
        Bth = 0.0
    if Bph != Bph:
        Bph = 0.0
        
    # Convert from spherical to ECEF
    Bx = Br*numpy.cos(lt)*numpy.cos(ln) + Bth*numpy.sin(lt)*numpy.cos(ln) - Bph*numpy.sin(ln)
    By = Br*numpy.cos(lt)*numpy.sin(ln) + Bth*numpy.sin(lt)*numpy.sin(ln) + Bph*numpy.cos(ln)
    Bz = Br*numpy.sin(lt) - Bth*numpy.cos(lt)
    
    # Are we done?
    if ecef:
        # For ECEF we don't need to do anything else
        outputField = Bx, By, Bz
        
    else:
        # Convert from ECEF to topocentric (geodetic)
        ## Update the coordinates for geodetic
        lt = lat*numpy.pi/180.0
        if 90.0 - lat < 0.001:
            lt = 89.999*numpy.pi/180.0
        elif 90.0 + lat < 0.001:
            lt = -89.999*numpy.pi/180.0
        else:
            lt = lat*numpy.pi/180.0
        ln = lng*numpy.pi/180.0
        
        ## Build the rotation matrix for ECEF to SEZ
        rot = numpy.array([[ numpy.sin(lt)*numpy.cos(ln), numpy.sin(lt)*numpy.sin(ln), -numpy.cos(lt)], 
                    [-numpy.sin(ln),               numpy.cos(ln),                0            ],
                    [ numpy.cos(lt)*numpy.cos(ln), numpy.cos(lt)*numpy.sin(ln),  numpy.sin(lt)]])
                    
        ## Apply and extract
        sez = numpy.dot(rot, numpy.array([Bx,By,Bz]))
        Bn, Be, Bz = -sez[0], sez[1], sez[2]
        
        outputField = Bn, Be, Bz
    
    # Done
    return outputField


def computeMagneticDeclination(Bn, Be, Bz):
    """
    Given the topocentric output of getMagneticField(), compute and return 
    the magnetic declination (deviation between magnetic north and true 
    north) in degrees.
    
    .. note::
        The declination is poorly defined (NaN) around the magnetic poles
        where the horizontal field strength is less than 100 nT.
    """
    
    # Compute the horizontal field strength
    Bh = numpy.sqrt(Bn**2+Be**2)
    
    # Compute the declination
    decl = 2.0*numpy.arctan2(Be, Bh+Bn)
    
    # Check for bounds
    if Bh < 100.0:
        decl = numpy.nan
        
    # Convert to degrees and done
    return decl*180.0/numpy.pi


def computeMagneticInclination(Bn, Be, Bz):
    """
    Given the topocentric output of getMagneticField(), compute and return 
    the magnetic inclination or magnetic dip (angle between the magnetic 
    field and the horizontal) in degrees.
    """
    
    # Compute the horizontal field strength
    Bh = numpy.sqrt(Bn**2+Be**2)
    
    # Compute the inclination.  This has an extra negative sign because of the
    # convention used in getMagneticField().
    incl = numpy.arctan2(-Bz, Bh)
    
    # Convert to degrees and done
    return incl*180.0/numpy.pi


def _downloadTECMapIGS(mjd, baseURL='ftp://cddis.gsfc.nasa.gov/gps/products/ionex/', timeout=120, type='final'):
    """
    Given an MJD value, download the corresponding IGS final data product 
    for that day.
    
    .. note::
        By default the "final" product is downloaded.  However, the "rapid" 
        data product may be downloaded if the 'type' keyword is set to 
        "rapid".
    """
    
    # Convert the MJD to a datetime instance so that we can pull out the year
    # and the day-of-year
    mpm = int((mjd - int(mjd))*24.0*3600.0*1000)
    dt = mjdmpm2datetime(int(mjd), mpm)
    
    year = dt.year
    dayOfYear = int(dt.strftime('%j'), 10)
    
    # Figure out which file we need to download
    if type == 'final':
        ## Final
        filename = 'igsg%03i0.%02ii.Z' % (dayOfYear, year%100)
    elif type == 'rapid':
        ## Rapid
        filename = 'igrg%03i0.%02ii.Z' % (dayOfYear, year%100)
    else:
        ## ???
        raise ValueError("Unknown TEC file type '%s'" % type)
        
    # Attempt to download the data
    try:
        tecFH = urlopen('%s/%04i/%03i/%s' % (baseURL, year, dayOfYear, filename), timeout=timeout)
        data = tecFH.read()
        tecFH.close()
    except IOError as e:
        __logger.error('Error downloading file for %i, %i: %s', dayOfYear, year, str(e))
        data = ''
    except socket.timeout:
        data = ''
        
    # Did we get anything?
    if len(data) == 0:
        ## Fail
        return False
    else:
        ## Success!  Save it to a file and then decompress it with 'gunzip' 
        ## since I can't figure out how to decompress this in Python.
        fh = open(os.path.join(_CacheDir, filename), 'wb')
        fh.write(data)
        fh.close()
        
        subprocess.check_call(['gunzip', '-f', os.path.join(_CacheDir, filename)])
        subprocess.check_call(['gzip', os.path.join(_CacheDir, os.path.splitext(filename)[0])])
        
        return True


def _downloadTECMapJPL(mjd, baseURL='ftp://cddis.gsfc.nasa.gov/gps/products/ionex/', timeout=120, type='final'):
    """
    Given an MJD value, download the corresponding JPL final data product 
    for that day.
    
    .. note::
        By default the "final" product is downloaded.  However, the "rapid" 
        data product may be downloaded if the 'type' keyword is set to 
        "rapid".
    """
    
    # Convert the MJD to a datetime instance so that we can pull out the year
    # and the day-of-year
    mpm = int((mjd - int(mjd))*24.0*3600.0*1000)
    dt = mjdmpm2datetime(int(mjd), mpm)
    
    year = dt.year
    dayOfYear = int(dt.strftime('%j'), 10)
    
    # Figure out which file we need to download
    if type == 'final':
        ## Final
        filename = 'jplg%03i0.%02ii.Z' % (dayOfYear, year%100)
    elif type == 'rapid':
        ## Rapid
        filename = 'jprg%03i0.%02ii.Z' % (dayOfYear, year%100)
    else:
        ## ???
        raise ValueError("Unknown TEC file type '%s'" % type)
        
    # Attempt to download the data
    try:
        tecFH = urlopen('%s/%04i/%03i/%s' % (baseURL, year, dayOfYear, filename), timeout=timeout)
        data = tecFH.read()
        tecFH.close()
    except IOError as e:
        __logger.error('Error downloading file for %i, %i: %s', dayOfYear, year, str(e))
        data = ''
    except socket.timeout:
        data = ''
        
    # Did we get anything?
    if len(data) == 0:
        ## Fail
        return False
    else:
        ## Success!  Save it to a file and then decompress it with 'gunzip' 
        ## since I can't figure out how to decompress this in Python.
        fh = open(os.path.join(_CacheDir, filename), 'wb')
        fh.write(data)
        fh.close()
        
        subprocess.check_call(['gunzip', '-f', os.path.join(_CacheDir, filename)])
        subprocess.check_call(['gzip', os.path.join(_CacheDir, os.path.splitext(filename)[0])])
        
        return True


def _downloadTECMapUQR(mjd, baseURL='ftp://cddis.gsfc.nasa.gov/gps/products/ionex/', timeout=120, type='final'):
    """
    Given an MJD value, download the corresponding JPL final data product 
    for that day.
    
    .. note::
        By default the "final" product is downloaded.  However, the "rapid" 
        data product may be downloaded if the 'type' keyword is set to 
        "rapid".
    """
    
    # Convert the MJD to a datetime instance so that we can pull out the year
    # and the day-of-year
    mpm = int((mjd - int(mjd))*24.0*3600.0*1000)
    dt = mjdmpm2datetime(int(mjd), mpm)
    
    year = dt.year
    dayOfYear = int(dt.strftime('%j'), 10)
    
    # Figure out which file we need to download
    if type == 'final':
        ## Final
        filename = 'uqrg%03i0.%02ii.Z' % (dayOfYear, year%100)
    elif type == 'rapid':
        ## Rapid
        filename = 'uqrg%03i0.%02ii.Z' % (dayOfYear, year%100)
    else:
        ## ???
        raise ValueError("Unknown TEC file type '%s'" % type)
        
    # Attempt to download the data
    try:
        tecFH = urlopen('%s/%04i/%03i/%s' % (baseURL, year, dayOfYear, filename), timeout=timeout)
        data = tecFH.read()
        tecFH.close()
    except IOError as e:
        __logger.error('Error downloading file for %i, %i: %s', dayOfYear, year, str(e))
        data = ''
    except socket.timeout:
        data = ''
        
    # Did we get anything?
    if len(data) == 0:
        ## Fail
        return False
    else:
        ## Success!  Save it to a file and then decompress it with 'gunzip' 
        ## since I can't figure out how to decompress this in Python.
        fh = open(os.path.join(_CacheDir, filename), 'wb')
        fh.write(data)
        fh.close()
        
        subprocess.check_call(['gunzip', '-f', os.path.join(_CacheDir, filename)])
        subprocess.check_call(['gzip', os.path.join(_CacheDir, os.path.splitext(filename)[0])])
        
        return True


def _downloadTECMapCODE(mjd, baseURL='ftp://ftp.aiub.unibe.ch/CODE/', timeout=120, type='final'):
    """
    Given an MJD value, download the corresponding CODE final data product 
    for that day.
    
    .. note::
        The 'type' keyword is ignored in the call.  It is included for 
        compatiability with _downloadTECMapIGS().
    """
    
    # Convert the MJD to a datetime instance so that we can pull out the year
    # and the day-of-year
    mpm = int((mjd - int(mjd))*24.0*3600.0*1000)
    dt = mjdmpm2datetime(int(mjd), mpm)
    
    year = dt.year
    dayOfYear = int(dt.strftime('%j'), 10)
    
    # Figure out which file we need to download
    filename = 'CODG%03i0.%02iI.Z' % (dayOfYear, year%100)
    
    # Attempt to download the data
    try:
        tecFH = urlopen('%s/%04i/%s' % (baseURL, year, filename), timeout=timeout)
        data = tecFH.read()
        tecFH.close()
    except IOError as e:
        __logger.error('Error downloading file for %i, %i: %s', dayOfYear, year, str(e))
        data = ''
    except socket.timeout:
        data = ''
        
    # Did we get anything?
    if len(data) == 0:
        ## Fail
        return False
    else:
        ## Success!  Save it to a file and then decompress it with 'gunzip' 
        ## since I can't figure out how to decompress this in Python.
        fh = open(os.path.join(_CacheDir, filename), 'wb')
        fh.write(data)
        fh.close()
        
        subprocess.check_call(['gunzip', '-f', os.path.join(_CacheDir, filename)])
        subprocess.check_call(['gzip', os.path.join(_CacheDir, os.path.splitext(filename)[0])])
        
        return True


def _downloadTECMapUSTEC(mjd, baseURL='http://www.ngdc.noaa.gov/stp/iono/ustec/products/', timeout=120):
    """
    Given an MJD value, download the corresponding JPL final data product 
    for that day.
    
    .. note::
        By default the "final" product is downloaded.  However, the "rapid" 
        data product may be downloaded if the 'type' keyword is set to 
        "rapid".
    """
    
    # Convert the MJD to a datetime instance so that we can pull out the year
    # and the day-of-year
    mpm = int((mjd - int(mjd))*24.0*3600.0*1000)
    dt = mjdmpm2datetime(int(mjd), mpm)
    
    year = dt.year
    month = dt.month
    dateStr = dt.strftime("%Y%m%d")
    # Build up the filename
    filename = '%s_ustec.tar.gz' % dateStr
    
    # Attempt to download the data
    try:
        tecFH = urlopen('%s/%04i/%02i/%s' % (baseURL, year, month, filename), timeout=timeout)
        data = tecFH.read()
        tecFH.close()
    except IOError as e:
        __logger.error('Error downloading file for %s: %s', dateStr, str(e))
        data = ''
    except socket.timeout:
        data = ''
        
    # Did we get anything?
    if len(data) == 0:
        ## Fail
        return False
    else:
        ## Success!  Save it to a file and then decompress it with 'gunzip' 
        ## since I can't figure out how to decompress this in Python.
        fh = open(os.path.join(_CacheDir, filename), 'wb')
        fh.write(data)
        fh.close()
        
        return True


def _parseTECMap(filename):
    """
    Given the name of a file containing a TEC map from the IGC, parse it 
    and return a dictionary containing the files data.
    
    The dictionary keys are:
    * dates - array of MJD values for each time step in the map
    * lats - 2-D array of latitude values for the maps in degrees
    * lngs - 2-D array of longitude values for the maps in degrees
    * height - height for the ionospheric pierce point in km
    * tec - 3-D array of TEC values in TECU.  The dimensions are time by
        latitude by longitude.
    * rms - 3-D array of TEC RMS values in TECU.  The dimensions are time
        by latitude by longitude.
    """
    
    # Variables to hold the map sequences
    dates = []
    tecMaps = []
    rmsMaps = []
    
    # State control variables to help keep up with where we are
    inMap = False
    inBlock = False
    
    # Go
    with gzip.open(filename, 'r') as fh:
        for line in fh:
            ## Are we beginning a map?
            line = line.replace('\n', '')
            if line.find('START OF TEC MAP') != -1 or line.find('START OF RMS MAP') != -1:
                inMap = True
                continue
                
            ## Have we just ended a map?
            if line.find('END OF TEC MAP') != -1 or line.find('END OF RMS MAP') != -1:
                if line.find('TEC') != -1:
                    tecMaps.append( cmap )
                else:
                    rmsMaps.append( cmap )
                
                inMap = False
                continue
                
            ## Are we in a map?
            if inMap:
                ## Is this part of the preamble? 
                if line.find('EPOCH OF CURRENT MAP') != -1:
                    ### Parse the date/time string
                    year, month, day, hour, minute, second = line.split(None, 6)[:6]
                    year = int(year)
                    month = int(month)
                    day = int(day)
                    hour = int(hour)
                    minute = int(minute)
                    second = int(second)
                    
                    ### Figure out the MJD
                    try:
                        dt = datetime(year, month, day, hour, minute, second, 0)
                    except ValueError:
                        if hour >= 24:
                            dt = datetime(year, month, day, hour-24, minute, second, 0)
                            dt += timedelta(days=1)
                        else:
                            continue
                    mjd, mpm = datetime2mjdmpm(dt)
                    mjd = mjd + mpm/1000.0/3600.0/24.0
                    if mjd not in dates:
                        dates.append( mjd )
                        
                    ### Initialize the map and the coorindates
                    cmap = []
                    lats = []
                    lngs = []
                    
                    continue
                    
                ## Is this a different part of the preamble? 
                elif line.find('LAT/LON1/LON2/DLON/H') != -1:
                    lat = float(line[3:8])
                    lng1 = float(line[8:14])
                    lng2 = float(line[14:20])
                    dlng = float(line[20:26])
                    height = float(line[26:32])
                    
                    cmap.append( [] )
                    lats.append( lat )
                    lngs = list(numpy.arange(lng1, lng2+dlng, dlng))
                    
                    inBlock = True
                    continue
                    
                ## Process the data block keeping in mind that missing values are stored 
                ## as 9999
                if inBlock:
                    fields = numpy.array([float(v)/10.0 for v in line.split(None)])
                    fields[numpy.where( fields == 999.9 )] = numpy.nan
                    
                    cmap[-1].extend( fields )
                    continue
                    
    # Combine everything together
    dates = numpy.array(dates, dtype=numpy.float64)
    tec = numpy.array(tecMaps, dtype=numpy.float32)
    rms = numpy.array(rmsMaps, dtype=numpy.float32)
    lats = numpy.array(lats, dtype=numpy.float32)
    lngs = numpy.array(lngs, dtype=numpy.float32)
    
    # Do we have a valid RMS map?  If not, make one.
    if rms.size != tec.size:
        rms = tec*0.05
        
    # Make lats and lngs 2-D to match the data
    lngs, lats = numpy.meshgrid(lngs, lats)
    
    # Build up the output
    output = {'dates': dates, 'lats': lats, 'lngs': lngs, 'height': height, 'tec': tec, 'rms': rms}
    
    # Done
    return output


def _parseUSTECMapIndividual(filename):
    """
    Parse an individual TEC map from the USTEC project.  This returns a five-
    element tuple of:
    * datetime for the start of the map
    * 2-D array of latitude values for the maps in degrees
    * 2-D array of longitude values for the maps in degrees
    * 2-D array of TEC values in TECU.  The dimensions are latitude by 
        longitude
    * 2-D array of TEC RMS values in TECU.  The dimensions are latitude 
        by longitude.
    
    Format Reference:
    https://www.ngdc.noaa.gov/stp/iono/ustec/README.html
    """
    
    # Get the date/time from the filename
    dt = os.path.basename(filename).split('_')[0]
    dt = datetime.strptime(dt, "%Y%m%d%H%M")
    
    # Open the TEC file for reading
    with open(filename, 'r') as fh:
        # Go!
        inBlock = False
        lats = []
        data = []
        for line in fh:
            if line[0] in ('#', ':'):
                ## Comments
                continue
            elif len(line) < 3:
                ## Blank lines
                continue
                
            ## Start the parsing
            if not inBlock:
                ### The first row consists of a list of longitudes
                fields = line.split()
                lngs = [float(f)/10.0 for f in fields[1:]]
                inBlock = True
                continue
            else:
                ### Slant TEC values are stored at the end of the file
                if line[:3] == '999':
                    inBlock = False
                    break
                    
                ### Before the satellite we have the TEC values, one for each latitude
                fields = line.split()
                lats.append( float(fields[0])/10 )
                data.append( [float(f)/10 for f in fields[1:]] )
                
    # Bring it into NumPy
    lats = numpy.array(lats)
    lngs = numpy.array(lngs)
    lngs,lats = numpy.meshgrid(lngs,lats)
    data = numpy.array(data)
    
    # Check for an associated RMS file
    rmsname = filename.replace('_TEC', '_ERR')
    if os.path.exists(rmsname):
        ## Oh good, we have one
        with open(rmsname, 'r') as fh:
            ## Go! (again)
            inBlock = False
            rlats = []
            rdata = []
            for line in fh:
                if line[0] in ('#', ':'):
                    ## Comments
                    continue
                elif len(line) < 3:
                    ## Blank lines
                    continue
                    
                ## Start the parsing
                if not inBlock:
                    ### The first row consists of a list of longitudes
                    fields = line.split()
                    rlngs = [float(f)/10.0 for f in fields[1:]]
                    inBlock = True
                    continue
                else:
                    ### Slant TEC values are stored at the end of the file
                    if line[:3] == '999':
                        inBlock = False
                        break
                        
                    ### Before the satellite we have the TEC RMS values, one for each latitude
                    fields = line.split()
                    rlats.append( float(fields[0])/10 )
                    rdata.append( [float(f)/10 for f in fields[1:]] )
                    
        # Bring it into NumPy
        rlats = numpy.array(rlats)
        rlngs = numpy.array(rlngs)
        rdata = numpy.array(rdata)
        
        # For some reason the RMS map has a lower resolution than the actual TEC map.
        # Interpolate up the the resolution of the actual TEC map so that we have an
        # uncertainty at each point
        interpFunction = RectBivariateSpline(rlats, rlngs, rdata, kx=1, ky=1)
        rms = data*0.0
        for i in xrange(lats.shape[0]):
            for j in xrange(lats.shape[0]):
                rms[i,j] = interpFunction(lats[i,j], lngs[i,j])
    else:
        ## Sadness, no RMS file found...
        rms = data*0.05
        
    # Reverse
    lats = lats[::-1,:]
    lngs = lngs[::-1,:]
    data = data[::-1,:]
    rms  = rms[::-1,:]
    
    # Done
    return dt, lats, lngs, data, rms


def _parseUSTECMapHeight(filename):
    """
    Parse emperical orthonormal functions to come up with an effective 
    height for the ionosphere.
    
    Format Reference:
    https://www.ngdc.noaa.gov/stp/iono/ustec/README.html
    """
    
    # Open the EOF file for reading
    with open(filename, 'r') as fh:
        # Go!
        inBlock = False
        heights = []
        data = []
        for line in fh:
            if line[0] in ('#', ':'):
                ## Comments
                continue
            elif len(line) < 3:
                ## Blank lines
                continue
                
            ## Start the parsing
            if not inBlock:
                fields = line.split()
                height = float(fields[2])
                step = float(fields[3])
                inBlock = True
            else:
                fields = line.split()
                heights.append( height - 6371)
                height += step
                data.append( float(fields[0]) )
                
    # Bring it into Numpy and get an average height
    heights = numpy.array(heights)
    data = numpy.array(data)
    height = (heights*data).sum() / data.sum()
    
    # Done
    return height


def _parseUSTECMap(filename):
    """
    Given the name of a file containing a TEC map from the USTEC project, 
    parse it and return a dictionary containing the files data.
    
    The dictionary keys are:
    * dates - array of MJD values for each time step in the map
    * lats - 2-D array of latitude values for the maps in degrees
    * lngs - 2-D array of longitude values for the maps in degrees
    * height - height for the ionospheric pierce point in km
    * tec - 3-D array of TEC values in TECU.  The dimensions are time by
        latitude by longitude.
    * rms - 3-D array of TEC RMS values in TECU.  The dimensions are time
        by latitude by longitude.
    """
    
    tempDir = tempfile.mkdtemp(prefix='ionosphere-')
    
    tf = tarfile.open(filename, 'r:*')
    tecFiles = [tio for tio in tf.getmembers() if tio.name.find('_TEC.txt') != -1]
    errFiles = [tio for tio in tf.getmembers() if tio.name.find('_ERR.txt') != -1]
    eofFiles = [tio for tio in tf.getmembers() if tio.name.find('_EOF.txt') != -1]
    tf.extractall(path=tempDir, members=tecFiles)
    tf.extractall(path=tempDir, members=errFiles)
    tf.extractall(path=tempDir, members=eofFiles)
    
    # Variables to hold the map sequences
    dates = []
    tecMaps = []
    rmsMaps = []
    
    # Get all of the TEC map files and load them in
    tecfilenames = glob.glob(os.path.join(tempDir, '*_TEC.txt'))
    tecfilenames.sort()
    for tecfilename in tecfilenames:
        #try:
        dt, lats, lngs, tec, rms = _parseUSTECMapIndividual(tecfilename)
        
        ### Figure out the MJD
        mjd, mpm = datetime2mjdmpm(dt)
        mjd = mjd + mpm/1000.0/3600.0/24.0
        if mjd not in dates:
            dates.append( mjd )
            
        # Stack on the new TEC and RMS maps
        tecMaps.append( tec )
        rmsMaps.append( rms )
            
        #except Exception as e:
            #pass
            
    # Get the mean ionospheric height
    eoffilename = glob.glob(os.path.join(tempDir, '*_EOF.txt'))[0]
    #try:
    height = _parseUSTECMapHeight(eoffilename)
    #except:
    #	height = 450
        
    # Cleanup
    tf.close()
    shutil.rmtree(tempDir, ignore_errors=True)
    
    # Combine everything together
    dates = numpy.array(dates, dtype=numpy.float64)
    tec = numpy.array(tecMaps, dtype=numpy.float32)
    rms = numpy.array(rmsMaps, dtype=numpy.float32)
    
    # Build up the output
    output = {'dates': dates, 'lats': lats, 'lngs': lngs, 'height': height, 'tec': tec, 'rms': rms}
    
    # Done
    return output


def _loadTECMap(mjd, timeout=120, type='IGS'):
    """
    Given an MJD value, load the corresponding TEC map.  If the map is not
    avaliable on disk, download it.
    """
    
    # Figure out which map to use
    if type == 'IGS':
        ## Cache entry name
        cacheName = 'TEC-IGS-%i' % mjd
        
        ## Download helper
        downloader = _downloadTECMapIGS
        
        ## Filename templates
        filenameTemplate = 'igsg%03i0.%02ii.gz'
        filenameAltTemplate = 'igrg%03i0.%02ii.gz'
        
    elif type == 'JPL':
        ## Cache entry name
        cacheName = 'TEC-JPL-%i' % mjd
        
        ## Download helper
        downloader = _downloadTECMapJPL
        
        ## Filename templates
        filenameTemplate = 'jplg%03i0.%02ii.gz'
        filenameAltTemplate = 'jprg%03i0.%02ii.gz'
        
    elif type == 'UQR':
        ## Cache entry name
        cacheName = 'TEC-UQR-%i' % mjd
        
        ## Download helper
        downloader = _downloadTECMapUQR
        
        ## Filename templates
        filenameTemplate = 'uqrg%03i0.%02ii.gz'
        filenameAltTemplate = 'uqrg%03i0.%02ii.gz'
        
    elif type == 'CODE':
        ## Cache entry name
        cacheName = 'TEC-CODE-%i' % mjd
        
        ## Download helper
        downloader = _downloadTECMapCODE
        
        ## Filename templates
        filenameTemplate = 'CODG%03i0.%02iI.gz'
        filenameAltTemplate = 'CODG%03i0.%02iI.gz'
        
    elif type == 'USTEC':
        ## Cache entry name
        cacheName = 'TEC-USTEC-%i' % mjd
        
        ## Download helper
        downloader = _downloadTECMapUSTEC
        
        ## Filename templates
        filenameTemplate = '%s_ustec.tar.gz'
        filenameAltTemplate = '%s_ustec.tar.gz'
        
    else:
        raise ValueError("Unknown data source '%s'" % type)
        
    try:
        # Is it already in the on-line cache?
        tecMap = _Cache[cacheName]
    except KeyError:
        # Nope, we need to fetch it
        
        # Convert the MJD to a datetime instance so that we can pull out the year
        # and the day-of-year
        mpm = int((mjd - int(mjd))*24.0*3600.0*1000)
        dt = mjdmpm2datetime(int(mjd), mpm)
        
        if type == 'USTEC':
            # Pull out a YMD string
            dateStr = dt.strftime("%Y%m%d")
            
            # Figure out the filenames in order of preference.  We'd rather have
            # final values than rapid values
            filename = filenameTemplate % (dateStr)
            
            # Is the primary file in the disk cache?
            if not os.path.exists(os.path.join(_CacheDir, filename)):
                ## Can we download it?
                status = downloader(mjd, timeout=timeout)
                
            else:
                ## Good we have the primary file
                pass
                
            # Parse it
            _Cache[cacheName] = _parseUSTECMap(os.path.join(_CacheDir, filename))
            
        else:
            
            # Pull out the year and the day-of-year
            year = dt.year
            dayOfYear = int(dt.strftime('%j'), 10)
            
            # Figure out the filenames in order of preference.  We'd rather have
            # final values than rapid values
            filename = filenameTemplate % (dayOfYear, year%100)
            filenameAlt = filenameAltTemplate % (dayOfYear, year%100)
            
            # Is the primary file in the disk cache?
            if not os.path.exists(os.path.join(_CacheDir, filename)):
                ## Can we download it?
                status = downloader(mjd, timeout=timeout, type='final')
                if not status:
                    ## Nope, now check for the secondary file on disk
                    if not os.path.exists(os.path.join(_CacheDir, filenameAlt)):
                        ## Can we download it?
                        status = downloader(mjd, timeout=timeout, type='rapid')
                        if status:
                            ### Good, we have the secondary file
                            filename = filenameAlt
                    else:
                        ### Good, we have the secondary file
                        filename = filenameAlt
            else:
                ## Good we have the primary file
                pass
                
            # Parse it
            _Cache[cacheName] = _parseTECMap(os.path.join(_CacheDir, filename))
            
        tecMap = _Cache[cacheName]
        
    # Done
    return tecMap


def getTECValue(mjd, lat=None, lng=None, includeRMS=False, timeout=120, type='IGS'):
    """
    Given an MJD value and, optionally, a latitude and longitude in degrees, 
    compute the TEC value in TECU above that location using data from the 
    IGS or CODE (depending on the value of the 'type' keyword).  If the 
    'includeRMS' keyword is set to True, also return the RMS in the TEC 
    value.
    """
    
    # Load in the right map
    tecMap = _loadTECMap(mjd, timeout=timeout, type=type)
    
    if type == 'USTEC':
        # Figure out the closest model point(s) to the requested MJD taking into
        # account that a new model is generated every fifteen minutes
        best = numpy.where( numpy.abs((tecMap['dates']-mjd)) < 15/60./24.0 )[0]
    else:
        # Figure out the closest model point(s) to the requested MJD taking into
        # account that a new model is generated every two hours
        best = numpy.where( numpy.abs((tecMap['dates']-mjd)) < 2/24.0 )[0]
        
    # Interpolate in time
    ## TEC
    slope = tecMap['tec'][best[-1],:,:] - tecMap['tec'][best[0],:,:]
    if best[-1] == best[0]:
        slope = 0.0
    else:
        slope /= tecMap['dates'][best[-1]] - tecMap['dates'][best[0]]
    tec = slope*(mjd - tecMap['dates'][best[0]]) + tecMap['tec'][best[0],:,:]
    
    ## RMS
    if includeRMS:
        slope = tecMap['rms'][best[-1],:,:] - tecMap['rms'][best[0],:,:]
        if best[-1] == best[0]:
            slope = 0.0
        else:
            slope /= tecMap['dates'][best[-1]] - tecMap['dates'][best[0]]
        rms = slope*(mjd - tecMap['dates'][best[0]]) + tecMap['rms'][best[0],:,:]
        
    # Interpolate in location, if desired
    if lat is not None and lng is not None:
        ## TEC
        interpFunction = RectBivariateSpline(tecMap['lats'][::-1,0], tecMap['lngs'][0,:], tec[::-1,:], kx=1, ky=1)
        tec = interpFunction(lat, lng)
        
        ## RMS
        if includeRMS:
            interpFunction = RectBivariateSpline(tecMap['lats'][::-1,0], tecMap['lngs'][0,:], rms[::-1,:], kx=1, ky=1)
            rms = interpFunction(lat, lng)
            
    # Done
    if includeRMS:
        return tec, rms
    else:
        return tec


def getIonosphericPiercePoint(site, az, el, height=450e3, verbose=False):
    """
    Given a site and a pointing direction (azimuth and elevation in degrees),
    compute the location of the ionospheric pierce  point.  Since the height
    assumed for the ionosphere is model-dependent the 'height' keyword sets 
    the elevation to use in meters.  Returns a three-element tuple of 
    latitude (degrees, North is positive), longitude (degrees, East is 
    positive), and elevation (meters).
    """
    
    # Create the various functions we need to figure out this optimization 
    # problem.  Maybe there is a better way to do this.
    
    ## Function that takes in a two-element tuple of latitude and longitude
    ## and returns the azimuth and elevation relative to the observer assuming
    ## a particular height.
    def func(params, xdata, site=site, elev=height):
        lat,lon = params
        
        az,el,d = site.getPointingAndDistance((lat, lon, elev))
        az %= (2*numpy.pi)
        
        az *= 180/numpy.pi
        el *= 180/numpy.pi
        
        return numpy.array([az, el])
        
    ## Error function that computes the difference between the input and the 
    ## model used in func().
    def err(params, ydata, xdata):
        return ydata - func(params, xdata)
        
    ## Secondary error function that computes a sum(err**2) that can be used
    ## with scipy.optimize.fmin()
    def err2(params, ydata, xdata):
        return (err(params, ydata, xdata)**2).sum()
        
    # Initial conditions for the optimization - we start directly overhead
    lat = site.lat * 180/numpy.pi
    lon = site.lon * 180/numpy.pi
    elev = site.elev + height
    x0 = (lat, lon)
    
    # Optimize
    output = fmin(err2, x0, args=(numpy.array([az, el]), []), disp=verbose)
    
    # Done
    return output[0], output[1], height



# -*- coding: utf-8 -*-

"""Unit test for the lsl.statistics.tests module."""

import os
import time
import unittest
import numpy

from lsl.statistics import stattests


__revision__ = "$Rev: 2394 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class stattests_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.statistics.tests
    module."""

    def test_waldwolfowitz(self):
        """Test the Wald-Wolfowitz (runs) test"""

        data = numpy.random.randn(1024)
        pValue = stattests.waldwolfowitz(data)


class stattests_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.statistics.tests 
    units tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(stattests_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for the lsl.writer.sdfits module."""

import os
import time
import unittest
import numpy
import tempfile
import pyfits

from lsl.common import stations as lwa_common
from lsl.writer import sdfits


__revision__ = "$Rev: 2675 $"
__version__  = "0.2"
__author__    = "Jayce Dowell"


class sdfits_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.writer.sdfits
    module."""

    testPath = None
    
    def setUp(self):
        """Turn off all numpy warnings and create the temporary file directory."""

        numpy.seterr(all='ignore')
        self.testPath = tempfile.mkdtemp(prefix='test-sdfits-', suffix='.tmp')

    def __initData(self):
        """Private function to generate a random set of data for writing a SDFITS 
        file.  The data is returned as a dictionary with keys:
        * freq - frequency array in Hz
        * site - lwa.common.stations object
        * stands - array of stand numbers
        * spec - array of spectrometer data stand x freq format
        """

        # Frequency range
        freq = numpy.arange(0,512)*20e6/512 + 40e6
        # Site and stands
        site = lwa_common.lwa1
        antennas = site.getAntennas()[0:40:2]
        
        # Set data
        specData = numpy.random.rand(len(antennas), len(freq))
        specData = specData.astype(numpy.float32)

        return {'freq': freq, 'site': site, 'antennas': antennas, 'spec': specData}
    
    def test_write_tables(self):
        """Test if the SDFITS writer writes all of the tables."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'sd-test-W.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = sdfits.SD(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.addComment('This is a comment')
        fits.addHistory('This is history')
        fits.addDataSet(testTime, 6.0, data['antennas'], data['spec'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        # Check that all of the extensions are there
        extNames = [hdu.name for hdu in hdulist]
        for ext in ['SINGLE DISH',]:
            self.assertTrue(ext in extNames)
        # Check the comments and history
        self.assertTrue('This is a comment' in str(hdulist[0].header['COMMENT']).split('\n'))
        self.assertTrue('This is history' in str(hdulist[0].header['HISTORY']).split('\n'))
        
        hdulist.close()
    
    def test_data(self):
        """Test the data table in the SINGLE DISH extension"""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'sd-test-data.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = sdfits.SD(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.addDataSet(testTime, 6.0, data['antennas'], data['spec'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        sd = hdulist['SINGLE DISH'].data

        # Correct number of elements
        self.assertEqual(len(sd.field('DATA')), len(data['antennas']))
        
        # Correct values
        for beam, spec in zip(sd.field('BEAM'), sd.field('DATA')):
            # Find out which visibility set in the random data corresponds to the 
            # current visibility
            i = 0
            for ant in data['antennas']:
                if ant.id == beam:
                    break
                else:
                    i = i + 1
            
            # Extract the data and run the comparison
            for fd, sd in zip(spec[0,0,0,:], data['spec'][i,:]):
                self.assertAlmostEqual(fd, sd, 8)
            i = i + 1
        
        hdulist.close()
    
    def tearDown(self):
        """Remove the test path directory and its contents"""

        tempFiles = os.listdir(self.testPath)
        for tempFile in tempFiles:
            os.unlink(os.path.join(self.testPath, tempFile))
        os.rmdir(self.testPath)
        self.testPath = None


class  sdfits_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.sim.vis units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)

        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(sdfits_tests)) 


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Plot the uv-plane coverage of LWA1 for a zenith snapshot and the expected 
beam.
"""

import sys
import math
import numpy
import argparse

from lsl.common import stations, metabundle, metabundleADP
from lsl.correlator import uvUtils
from lsl.misc import parser as aph

import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter,  MaxNLocator


def main(args):
    # Setup the LWA station information
    if args.metadata is not None:
        try:
            station = stations.parseSSMIF(args.metadata)
        except ValueError:
            try:
                station = metabundle.getStation(args.metadata, ApplySDM=True)
            except:
                station = metabundleADP.getStation(args.metadata, ApplySDM=True)
    elif args.lwasv:
        station = stations.lwasv
    else:
        station = stations.lwa1
        
    antennas = []
    for ant in station.getAntennas()[0::2]:
        if ant.getStatus() == 33:
            antennas.append(ant)
    print "Displaying uv coverage for %i good stands" % len(antennas)
    
    HA = 0.0
    dec = station.lat*180.0/math.pi
    
    uvw = uvUtils.computeUVW(antennas, HA=HA, dec=dec, freq=args.frequency)
    uvw = numpy.squeeze(uvw[:,:,0])
    
    # Coursely grid the uv data to come up with a rough beam
    grid = numpy.zeros((1*240,1*240))
    for i in range(uvw.shape[0]):
        u = round((uvw[i,0]+120)*1)
        v = round((uvw[i,1]+120)*1)
        try:
            grid[u,v] += 1
        except IndexError:
            pass
            
    # Plot
    # Part 1 - Setup
    fig = plt.figure(figsize=(8,8))
    ax1 = plt.axes([0.30, 0.30, 0.60, 0.60])
    ax2 = plt.axes([0.30, 0.05, 0.60, 0.15])
    ax3 = plt.axes([0.05, 0.30, 0.15, 0.60])
    ax4 = plt.axes([0.08, 0.08, 0.15, 0.15])
    ax5 = plt.axes([0.32, 0.32, 0.15, 0.15])
    
    # Part 2 - Beam response (in dB)
    beam = numpy.fft.fft2(grid)
    beam = numpy.fft.fftshift(beam)
    beam = numpy.abs(beam)**2
    beam = numpy.log10(beam)*10.0
    ax5.imshow(beam[40:200,40:200], interpolation="nearest", vmin=numpy.median(beam), vmax=beam.max())
    ax5.xaxis.set_major_formatter( NullFormatter() )
    ax5.yaxis.set_major_formatter( NullFormatter() )
    
    # Part 3 - uv plane plot
    c = ax1.scatter(uvw[:,0], uvw[:,1], c=uvw[:,2], s=10.0, alpha=0.75)
    d = ax1.scatter(-uvw[:,0], -uvw[:,1], c=-uvw[:,2], s=10.0, alpha=0.75)
    ax1.set_xlabel('u [$\\lambda$]')
    ax1.set_ylabel('v [$\\lambda$]')
    ax1.set_title('UV Coverage for HA=%+.3f$^h$, $\delta$=%+.3f$^\circ$ at %s' % (HA, dec, station.name))
    
    # Part 4 - uw plane plot
    ax2.scatter(uvw[:,0], uvw[:,2], c=uvw[:,2], s=10.0)
    ax2.scatter(-uvw[:,0], -uvw[:,2], c=-uvw[:,2], s=10.0)
    ax2.xaxis.set_major_formatter( NullFormatter() )
    ax2.set_ylabel('w [$\\lambda$]')
    
    # Part 5 - wv plane plot
    ax3.scatter(uvw[:,2], uvw[:,1], c=uvw[:,2], s=10.0)
    ax3.scatter(-uvw[:,2], -uvw[:,1], c=-uvw[:,2], s=10.0)
    ax3.yaxis.set_major_formatter( NullFormatter() )
    ax3.set_xlabel('w [$\\lambda$]')
    
    # Part 6 - Histogram of uvw distances in lambda
    rad = numpy.zeros(uvw.shape[0])
    for i in range(rad.shape[0]):
        rad[i] = math.sqrt( uvw[i,0]**2.0 + uvw[i,1]**2.0 + uvw[i,2]**2.0 )
    try:
        ax4.hist(rad, 20)
    except TypeError:
        ## I don't know why this happens
        pass
    ax4.set_xlabel('uvw Radius [$\lambda$]')
    ax4.set_ylabel('Baselines')
    
    # Plot adjustment
    xlim = ax1.get_xlim()
    ylim = ax1.get_ylim()
    ax1.set_xlim([numpy.floor(xlim[0]/25.0)*25.0, numpy.ceil(xlim[1]/25.0)*25.0])
    ax1.set_ylim([numpy.floor(ylim[0]/25.0)*25.0, numpy.ceil(ylim[1]/25.0)*25.0])
    
    ax2.set_xlim( ax1.get_xlim() )
    ax2.yaxis.set_major_locator( MaxNLocator(nbins=4) )
    
    ax3.set_ylim( ax1.get_ylim() )
    ax3.xaxis.set_major_locator( MaxNLocator(nbins=4) )
    
    xlim = ax4.get_xlim()
    ylim = ax4.get_ylim()
    ax4.set_xlim([numpy.floor(xlim[0]/25.0)*25.0, numpy.ceil(xlim[1]/25.0)*25.0])
    ax4.set_ylim([numpy.floor(ylim[0]/5.e3)*5.e3, numpy.ceil(ylim[1]/5.e3)*5.e3])
    ax4.xaxis.set_major_locator( MaxNLocator(nbins=4) )
    ax4.yaxis.set_major_locator( MaxNLocator(nbins=4) )
    
    # Show n' save
    plt.show()
    if args.output is not None:
        fig.savefig(args.output)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='plot the UV-plane converage of an LWA station', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('-s', '--lwasv', action='store_true', 
                        help='use LWA-SV instead of LWA1')
    parser.add_argument('-f', '--frequency', type=aph.frequency, default='50.0', 
                        help='frequency in MHz to compute the uv coverage')
    parser.add_argument('-m', '--metadata', type=str, 
                        help='name of the SSMIF or metadata tarball file to use for mappings')
    parser.add_argument('-o', '--output', type=str, 
                        help='filename to save the plot to')
    args = parser.parse_args()
    main(args)
    



# -*- coding: utf-8 -*-

"""
Module for querying the earth orientation parameters for a given date/list
of dates.

.. versionchanged:: 1.0.3
    Added a fallback to the backup MAIA server 'toshi.nofs.navy.mil' if the
    primary server at 'toshi.nofs.navy.mil' cannot be reached

.. versionchanged:: 1.0.0
    Added caching of MAIA results to speed up subsequent calls to getEOP().
    Removed getEOPRange() since getEOP() can do the same thing.
"""

import os
import time
import ephem
import numpy
import socket
import logging
try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen
    
import lsl.astro as astro
from lsl.misc.total_sorting import cmp_to_total


__version__ = '0.3'
__revision__ = '$Rev: 2567 $'
__all__ = ['EOP', 'getEOP', '__version__', '__revision__', '__all__']


import warnings
warnings.warn("lsl.misc.geodesy will be removed in LSL version 1.3.0", DeprecationWarning)


# Logger for capturing problems with downloading EOP data
__logger = logging.getLogger('__main__')


# Create the cache directory
if not os.path.exists(os.path.join(os.path.expanduser('~'), '.lsl')):
    os.mkdir(os.path.join(os.path.expanduser('~'), '.lsl'))
_CacheDir = os.path.join(os.path.expanduser('~'), '.lsl', 'geodesy_cache')
if not os.path.exists(_CacheDir):
    os.mkdir(_CacheDir)


@cmp_to_total
class EOP(object):
    """
    Object for storing the geodetic parameters relevant for DiFX input 
    files:
    * mjd - modified Julian Date of the measurement/prediction
    * x - difference between the celestial ephemeric pole (CEP) and the 
        international reference pole (IRP) in the direction of the IERS
        meridian [arc seconds]
    * y - difference between the CEP and IRP in the direction of 90 degrees 
        west longitude [arc seconds]
    * dx - dX with respect to IAU2000A Nutation [arc seconds]
    * dy - dY with respect to IAU2000A Nutation [arc seconds]
    * UT1-UTC - difference between rotation angle about the pole and UTC
        [seconds]
    * type - whether the values for the given MJD are observed (final/IERS) or
        predicted.
        
    .. versionchanged:: 1.0.0
        Added extra attributes to store dX (dx) and dY (dy) in arc seconds.
    """
    
    def __init__(self, mjd=0.0, x=0.0, y=0.0, dx=0.0, dy=0.0, utDiff=0.0, type='final'):
        self.mjd = mjd
        self.x = x
        self.y = y
        self.dx = dx
        self.dy = dy
        self.utDiff = utDiff
        self.type = type
        
        self.date = None
        
        self.__setDate()
        
    def fromMAIA(self, line):
        """
        Given a line from a MAIA standard rapid EOP data (IAU2000) file, fill
        in the object's values with the needed information.
        """
        
        self.mjd = float(line[7:15])
        self.x = float(line[18:27])
        self.y = float(line[37:46])
        self.dx = float(line[97:106]) / 1000.0
        self.dy = float(line[116:125]) / 1000.0
        self.utDiff = float(line[58:68])
        if line[57] == 'I':
            self.type = 'final'
        else:
            self.type = 'prediction'
            
        self.__setDate()
        
    def __setDate(self):
        """
        Use the ephem.Data object to get an easy-to-use date into the structure.
        """
        
        self.date = ephem.Date(self.mjd + astro.MJD_OFFSET - astro.DJD_OFFSET)
        
    def __str__(self):
        """
        Create a string representation of the EOP object that shows the MJD, x, 
        y, and UT1-UTC values.
        """
        
        return "%.1f (%s): x=%.6f y=%.6f UT1-UTC=%.6f (%s)" % (self.mjd, str(self.date), self.x, self.y, self.utDiff, self.type)
        
    def __eq__(self, y):
        """
        Determine if MJDs of two EOP objects are equal, or if the MJD of a EOP 
        object equal data of a numeric MJD.
        """
        
        tX = self.mjd
        try:
            tY = y.mjd
        except:
            tY = float(y)
            
        if tX == tY:
            return True
        else:
            return False
            
    def __cmp__(self, y):
        """
        Method for soring EOP objects based on their MJDs.
        """
        
        tX = float(self.date)
        try:
            tY = float(y.date)
        except AttributeError:
            tY = float(y)
        if tY > tX:
            return -1
        elif tX > tY:
            return 1
        else:
            return 0


def __downloadFile(filename, baseURL='http://toshi.nofs.navy.mil/ser7/', timeout=120):
    try:
        eopFH = urlopen('%s%s' % (baseURL, filename), timeout=timeout)
        data = eopFH.read()
        eopFH.close()
    except IOError as e:
        __logger.error('Error downloading file \'%s\': %s', filename, str(e))
        data = ''
    except socket.timeout:
        __logger.error('Timeout after %i seconds downloading file \'%s\'', timeout, filename)
        data = ''
        
    if len(data) == 0:
        return False
    else:
        fh = open(os.path.join(_CacheDir, filename), 'wb')
        fh.write(data)
        fh.close()
        return True


def __loadHistoric1973(timeout=120):
    """
    Load in historical values from the web.  The downloaded file includes 
    values from January 2, 1973 until today (usually).
    """
    
    if not os.path.exists(os.path.join(_CacheDir, 'finals2000A.all')):
        status = __downloadFile('finals2000A.all', timeout=timeout)
        if not status:
            __downloadFile('finals2000A.all', baseURL='ftp://ftp.iers.org/products/eop/rapid/standard/', timeout=timeout)
    else:
        age = time.time() - os.stat(os.path.join(_CacheDir, 'finals2000A.all')).st_mtime
        if age > (3600*24*180):
            status = __downloadFile('finals2000A.all', timeout=timeout)
            if not status:
                __downloadFile('finals2000A.all', baseURL='ftp://ftp.iers.org/products/eop/rapid/standard/', timeout=timeout)
                
    eops = []
    with open(os.path.join(_CacheDir, 'finals2000A.all'), 'r') as fh:
        for line in fh:
            newEOP = EOP()
            try:
                newEOP.fromMAIA(line) 
                # Only include "final" values, not predictions
                if newEOP.type == 'final':
                    eops.append(newEOP)
            except:
                pass
    if len(eops) == 0:
        eops.append(None)
        
    return eops


def __loadHistoric1992(timeout=120):
    """
    Load in historical values from the web.  The downloaded file includes 
    values from January 1, 1992 until today (usually).
    """
    
    if not os.path.exists(os.path.join(_CacheDir, 'finals2000A.data')):
        status = __downloadFile('finals2000A.data', timeout=timeout)
        if not status:
            __downloadFile('finals2000A.data', baseURL='ftp://ftp.iers.org/products/eop/rapid/standard/', timeout=timeout)
    else:
        age = time.time() - os.stat(os.path.join(_CacheDir, 'finals2000A.data')).st_mtime
        if age > (3600*24*7):
            status = __downloadFile('finals2000A.data', timeout=timeout)
            if not status:
                __downloadFile('finals2000A.data', baseURL='ftp://ftp.iers.org/products/eop/rapid/standard/', timeout=timeout)
                
    eops = []
    with open(os.path.join(_CacheDir, 'finals2000A.data'), 'r') as fh:
        for line in fh:
            newEOP = EOP()
            try:
                newEOP.fromMAIA(line) 
                # Only include "final" values, not predictions
                if newEOP.type == 'final':
                    eops.append(newEOP)
            except:
                pass
    if len(eops) == 0:
        eops.append(None)
        
    return eops


def __loadCurrent90(timeout=120):
    """
    Load data for the current 90-day period from the web.
    """

    if not os.path.exists(os.path.join(_CacheDir, 'finals2000A.daily')):
        status = __downloadFile('finals2000A.daily', timeout=timeout)
        if not status:
            __downloadFile('finals2000A.daily', baseURL='ftp://ftp.iers.org/products/eop/rapid/daily/', timeout=timeout)
    else:
        age = time.time() - os.stat(os.path.join(_CacheDir, 'finals2000A.daily')).st_mtime
        if age > (3600*24):
            status = __downloadFile('finals2000A.daily', timeout=timeout)
            if not status:
                __downloadFile('finals2000A.daily', baseURL='ftp://ftp.iers.org/products/eop/rapid/daily/', timeout=timeout)
                
    eops = []
    with open(os.path.join(_CacheDir, 'finals2000A.daily'), 'r') as fh:
        for line in fh:
            newEOP = EOP()
            try:
                newEOP.fromMAIA(line) 
                eops.append(newEOP)
            except:
                pass
    if len(eops) == 0:
        eops.append(None)
        
    return eops


def getEOP(mjd=None, timeout=120):
    """
    Return a list of earth orientation parameter objects for the specified 
    MJDs.  A MJD of 'None' returns the values for today's date.
    
    .. versionchanged:: 0.5.2
        Added the `timeout' keyword to deal with failures download EOP data.
        The default value is 120 seconds.
        
    .. versionchanged:: 1.0.0
        Added caching of EOP values to speed up subsequent calls.
    """
    
    try:
        len(mjd)
    except TypeError:
        if mjd is None:
            try:
                mjd = [int(float(ephem.now()) + astro.DJD_OFFSET - astro.MJD_OFFSET)]
            except AttributeError:
                mjd = [int(float(ephem.now()) + 2415020.0 - astro.MJD_OFFSET)]
        else:
            mjd = [int(mjd)]
    mjd = numpy.array(mjd).astype(numpy.int32)

    oldEOPs = []
    midEOPs = []
    newEOPs = __loadCurrent90(timeout=timeout)
    if mjd.min() < 48622:
        oldEOPs = __loadHistoric1973(timeout=timeout)
    if mjd.min() < newEOPs[0].mjd:
        midEOPs = __loadHistoric1992(timeout=timeout)
        
    outEOPs = []
    for day in mjd:
        if day in newEOPs:
            outEOPs.append(newEOPs[newEOPs.index(day)])
        elif day in midEOPs:
            outEOPs.append(midEOPs[midEOPs.index(day)])
        elif day in oldEOPs:
            outEOPs.append(oldEOPs[oldEOPs.index(day)])
        else:
            outEOPs.append(None)
    
    if len(mjd) == 1:
        outEOPs = outEOPs[0]
    return outEOPs



# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range

"""
Module for writing correlator output to a CASA measurement set.

.. versionadded:: 1.2.1
"""

import os
import gc
import re
import glob
import math
import ephem
import numpy
import shutil
from datetime import datetime

from lsl import astro
from lsl.misc import geodesy
from lsl.common import constants
from lsl.misc.total_sorting import cmp_to_total

try:
    from collections import OrderedDict
except ImportError:
    from lsl.misc.OrderedDict import OrderedDict


__version__ = '0.1'
__revision__ = '$Rev: 2740 $'
__all__ = ['MS', 'StokesCodes', 'NumericStokes', 
           '__version__', '__revision__', '__all__']


StokesCodes = {'I':  1, 'Q':  2, 'U':  3, 'V':  4, 
               'RR': 5, 'RL': 6, 'LR': 7, 'LL': 8,
               'XX': 9, 'XY':10, 'YX':11, 'YY':12}
               

NumericStokes = { 1:'I',   2:'Q',   3:'U',   4:'V', 
                  5:'RR',  6:'RL',  7:'LR',  8:'LL',
                  9:'XX', 10:'XY', 11:'YX', 12:'YY'}

def mergeBaseline(ant1, ant2, shift=16):
    """
    Merge two stand ID numbers into a single baseline using the specified bit 
    shift size.
    """
    
    return (ant1 << shift) | ant2

def splitBaseline(baseline, shift=16):
    """
    Given a baseline, split it into it consistent stand ID numbers.
    """
    
    part = 2**shift - 1
    return (baseline >> shift) & part, baseline & part


try:
    from casacore.tables import table, tableutil
    
    class MS(object):
        """
        Class for storing visibility data and writing the data, along with array
        geometry, frequency setup, etc., to a CASA measurement set.
        """
        
        class _Antenna(object):
            """
            Holds information describing the location and properties of an antenna.
            """
            
            def __init__(self, id, x, y, z, bits=8):
                self.id = id
                self.x = x
                self.y = y
                self.z = z
                self.levels = bits
                self.polA = {'Type': 'X', 'Angle': 0.0, 'Cal': [0.0, 0.0]}
                self.polB = {'Type': 'Y', 'Angle': 90.0, 'Cal': [0.0, 0.0]}
                
            def getName(self):
                return "LWA%03i" % self.id
                
        class _Frequency:
            """
            Holds information about the frequency setup used in the file.
            """

            def __init__(self, offset, channelWidth, bandwidth):
                self.id = 1
                self.bandFreq = offset
                self.chWidth = channelWidth
                self.totalBW = bandwidth
                self.sideBand = 1
                self.baseBand = 0
                
        @cmp_to_total
        class _UVData(object):
            """
            Represents one UV visibility data set for a given observation time.
            """
        
            def __init__(self, obsTime, intTime, baselines, visibilities, pol=StokesCodes['XX'], source='z'):
                self.obsTime = obsTime
                self.intTime = intTime
                self.baselines = baselines
                self.visibilities = visibilities
                self.pol = pol
                self.source = source
                
            def __cmp__(self, y):
                """
                Function to sort the self.data list in order of time and then 
                polarization code.
                """
                
                sID = (self.obsTime, abs(self.pol))
                yID = (y.obsTime,    abs(y.pol)   )
                
                if sID > yID:
                    return 1
                elif sID < yID:
                    return -1
                else:
                    return 0
                    
            def time(self):
                return self.obsTime
                
            def getUVW(self, HA, dec, obs):
                Nbase = len(self.baselines)
                uvw = numpy.zeros((Nbase,3), dtype=numpy.float32)
                
                # Phase center coordinates
                # Convert numbers to radians and, for HA, hours to degrees
                HA2 = HA * 15.0 * numpy.pi/180
                dec2 = dec * numpy.pi/180
                lat2 = obs.lat
                
                # Coordinate transformation matrices
                trans1 = numpy.matrix([[0, -numpy.sin(lat2), numpy.cos(lat2)],
                                [1,  0,               0],
                                [0,  numpy.cos(lat2), numpy.sin(lat2)]])
                trans2 = numpy.matrix([[ numpy.sin(HA2),                  numpy.cos(HA2),                 0],
                                [-numpy.sin(dec2)*numpy.cos(HA2),  numpy.sin(dec2)*numpy.sin(HA2), numpy.cos(dec2)],
                                [ numpy.cos(dec2)*numpy.cos(HA2), -numpy.cos(dec2)*numpy.sin(HA2), numpy.sin(dec2)]])
                        
                for i,(a1,a2) in enumerate(self.baselines):
                    # Go from a east, north, up coordinate system to a celestial equation, 
                    # east, north celestial pole system
                    xyzPrime = a1.stand - a2.stand
                    xyz = trans1*numpy.matrix([[xyzPrime[0]],[xyzPrime[1]],[xyzPrime[2]]])
                    
                    # Go from CE, east, NCP to u, v, w
                    temp = trans2*xyz
                    uvw[i,:] = numpy.squeeze(temp)
                    
                return uvw
                    
            def argsort(self, mapper=None, shift=16):
                packed = []
                for a1,a2 in self.baselines:
                    if mapper is None:
                        s1, s2 = a1.stand.id, a2.stand.id
                    else:
                        s1, s2 = mapper.index(a1.stand.id), mapper.index(a2.stand.id)
                    packed.append( mergeBaseline(s1, s2, shift=shift) )
                packed = numpy.array(packed, dtype=numpy.int32)
                
                return numpy.argsort(packed)
                
        def parseRefTime(self, refTime):
            """
            Given a time as either a integer, float, string, or datetime object, 
            convert it to a string in the formation 'YYYY-MM-DDTHH:MM:SS'.
            """
            
            # Valid time string (modulo the 'T')
            timeRE = re.compile(r'\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}(\.\d+)?')
            
            if type(refTime) in (int, long, float):
                refDateTime = datetime.utcfromtimestamp(refTime)
                refTime = refDateTime.strftime("%Y-%m-%dT%H:%M:%S")
            elif type(refTime) == datetime:
                refTime = refTime.strftime("%Y-%m-%dT%H:%M:%S")
            elif type(refTime) == str:
                # Make sure that the string times are of the correct format
                if re.match(timeRE, refTime) is None:
                    raise RuntimeError("Malformed date/time provided: %s" % refTime)
                else:
                    refTime = refTime.replace(' ', 'T', 1)
            else:
                raise RuntimeError("Unknown time format provided.")
                
            return refTime
            
        def refTime2AstroDate(self):
            """
            Convert a reference time string to an :class:`lsl.astro.date` object.
            """
            
            dateStr = self.refTime.replace('T', '-').replace(':', '-').split('-')
            return astro.date(int(dateStr[0]), int(dateStr[1]), int(dateStr[2]), int(dateStr[3]), int(dateStr[4]), float(dateStr[5]))
            
        def __init__(self, filename, refTime=0.0, verbose=False, memmap=None, clobber=False):
            """
            Initialize a new FITS IDI object using a filename and a reference time 
            given in seconds since the UNIX 1970 ephem, a python datetime object, or a 
            string in the format of 'YYYY-MM-DDTHH:MM:SS'.
            
            .. versionchanged:: 1.1.2
                Added the 'memmap' and 'clobber' keywords to control if the file
                is memory mapped and whether or not to overwrite an existing file, 
                respectively.
            """
            
            # File-specific information
            self.filename = filename
            self.verbose = verbose
            
            # Observatory-specific information
            self.siteName = 'Unknown'
            
            # Observation-specific information
            self.refTime = self.parseRefTime(refTime)
            self.nAnt = 0
            self.nChan = 0
            self.nStokes = 0
            self.refVal = 0
            self.refPix = 0
            self.channelWidth = 0
            
            # Parameters that store the meta-data and data
            self.array = []
            self.freq = []
            self.stokes = []
            self.data = []
            
            # Open the file and get going
            if os.path.exists(filename):
                if clobber:
                    shutil.rmtree(filename, ignore_errors=False)
                else:
                    raise IOError("File '%s' already exists" % filename)
            self.basename = filename
            
        def setStokes(self, polList):
            """
            Given a list of Stokes parameters, update the object's parameters.
            """
            
            for pol in polList:
                if type(pol) == str:
                    numericPol = StokesCodes[pol.upper()]
                else:
                    numericPol = pol
                    
                if numericPol not in self.stokes:
                    self.stokes.append(numericPol)
                    
            # Sort into order of 'XX', 'YY', 'XY', and 'YX' or 'I', 'Q', 'U', and 'V'
            self.stokes.sort()
            if self.stokes[0] < 0:
                self.stokes.reverse()
                
            self.nStokes = len(self.stokes)
            
        def setFrequency(self, freq):
            """
            Given a numpy array of frequencies, set the relevant common observation
            parameters and add an entry to the self.freq list.
            """
            
            if self.nChan == 0:
                self.nChan = len(freq)
                self.refVal = freq[0]
                self.refPix = 1
                self.channelWidth = numpy.abs(freq[1] - freq[0])
                offset = 0.0
            else:
                assert(len(freq) == self.nChan)
                offset = freq[0] - self.refVal
            totalWidth = numpy.abs(freq[-1] - freq[0])
            
            freqSetup = self._Frequency(offset, self.channelWidth, totalWidth)

            self.freq.append(freqSetup)
            
        def setGeometry(self, site, antennas, bits=8):
            """
            Given a station and an array of stands, set the relevant common observation
            parameters and add entries to the self.array list.
            
            .. versionchanged:: 0.4.0
                Switched over to passing in Antenna instances generated by the
                :mod:`lsl.common.stations` module instead of a list of stand ID
                numbers.
            """
            
            # Update the observatory-specific information
            self.siteName = site.name
            
            stands = []
            for ant in antennas:
                stands.append(ant.stand.id)
            stands = numpy.array(stands)
            
            arrayX, arrayY, arrayZ = site.getGeocentricLocation()
            
            xyz = numpy.zeros((len(stands),3))
            for i,ant in enumerate(antennas):
                xyz[i,0] = ant.stand.x
                xyz[i,1] = ant.stand.y
                xyz[i,2] = ant.stand.z
                
            # Create the stand mapper
            mapper = []
            ants = []
            topo2eci = site.getECITransform()
            for i in xrange(len(stands)):
                eci = numpy.dot(topo2eci, xyz[i,:])
                ants.append( self._Antenna(stands[i], eci[0], eci[1], eci[2], bits=bits) )
                mapper.append( stands[i] )
                
            self.nAnt = len(ants)
            self.array.append( {'center': [arrayX, arrayY, arrayZ], 'ants': ants, 'mapper': mapper, 'inputAnts': antennas} )
            
        def addDataSet(self, obsTime, intTime, baselines, visibilities, pol='XX', source='z'):
            """
            Create a UVData object to store a collection of visibilities.
            
            .. versionchanged:: 0.4.0
                Switched over to passing in Antenna instances generated by the
                :mod:`lsl.common.stations` module instead of a list of stand ID
                as part of the baselines.
                
            .. versionchanged:: 1.1.0
                Added a new 'source' keyword to set the phase center for the data.
                This can either by 'z' for zenith or a ephem.Body instances for a
                point on the sky.
            """
            
            if type(pol) == str:
                numericPol = StokesCodes[pol.upper()]
            else:
                numericPol = pol
                
            self.data.append( self._UVData(obsTime, intTime, baselines, visibilities, pol=numericPol, source=source) )
            
        def write(self):
            """
            Fill in the FITS-IDI file will all of the tables in the 
            correct order.
            """
            
            # Validate
            if self.nStokes == 0:
                raise RuntimeError("No polarization setups defined")
            if len(self.freq) == 0:
                raise RuntimeError("No frequency setups defined")
            if self.nAnt == 0:
                raise RuntimeError("No array geometry defined")
            if len(self.data) == 0:
                raise RuntimeError("No visibility data defined")
                
            # Sort the data set
            self.data.sort()
            
            # Write the tables
            self._writeMain()
            self._writeAntenna()
            self._writePolarization()
            self._writeObservation()
            self._writeSpectralWindow()
            self._writeMisc()
            
            # Fixup the info and keywords for the main table
            tb = table("%s" % self.basename, readonly=False, ack=False)
            tb.putinfo({'type':'Measurement Set', 
                        'readme':'This is a MeasurementSet Table holding measurements from a Telescope'})
            tb.putkeyword('MS_VERSION', numpy.float32(2.0))
            for filename in sorted(glob.glob('%s/*' % self.basename)):
                if os.path.isdir(filename):
                    tname = os.path.basename(filename)
                    stb = table("%s/%s" % (self.basename, tname), ack=False)
                    tb.putkeyword(tname, stb)
                    stb.close()
            tb.flush()
            tb.close()
                
            # Clear out the data section
            del(self.data[:])
            gc.collect()
            
        def close(self):
            """
            Close out the file.
            """
            
            pass
            
        def _writeAntenna(self):
            """
            Write the antenna table.
            """
            
            col1 = tableutil.makearrcoldesc('OFFSET', 0.0, 1, 
                                            comment='Axes offset of mount to FEED REFERENCE point', 
                                            keywords={'QuantumUnits':['m','m','m'], 
                                                      'MEASINFO':{'type':'position', 'Ref':'ITRF'}
                                            })
            col2 = tableutil.makearrcoldesc('POSITION', 0.0, 1,
                                            comment='Antenna X,Y,Z phase reference position', 
                                            keywords={'QuantumUnits':['m','m','m'], 
                                                      'MEASINFO':{'type':'position', 'Ref':'ITRF'}
                                                      })
            col3 = tableutil.makescacoldesc('TYPE', "ground-based", 
                                            comment='Antenna type (e.g. SPACE-BASED)')
            col4 = tableutil.makescacoldesc('DISH_DIAMETER', 2.0, 
                                            comment='Physical diameter of dish', 
                                            keywords={'QuantumUnits':['m',]})
            col5 = tableutil.makescacoldesc('FLAG_ROW', False, 
                                            comment='Flag for this row')
            col6 = tableutil.makescacoldesc('MOUNT', "alt-az", 
                                            comment='Mount type e.g. alt-az, equatorial, etc.')
            col7 = tableutil.makescacoldesc('NAME', "none", 
                                            comment='Antenna name, e.g. VLA22, CA03')
            col8 = tableutil.makescacoldesc('STATION', self.siteName, 
                                            comment='Station (antenna pad) name')
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7, col8])
            tb = table("%s/ANTENNA" % self.basename, desc, nrow=self.nAnt, ack=False)
            
            tb.putcol('OFFSET', numpy.zeros((self.nAnt,3)), 0, self.nAnt)
            tb.putcol('TYPE', ['GROUND-BASED,']*self.nAnt, 0, self.nAnt)
            tb.putcol('DISH_DIAMETER', [2.0,]*self.nAnt, 0, self.nAnt)
            tb.putcol('FLAG_ROW', [False,]*self.nAnt, 0, self.nAnt)
            tb.putcol('MOUNT', ['ALT-AZ',]*self.nAnt, 0, self.nAnt)
            tb.putcol('NAME', [ant.getName() for ant in self.array[0]['ants']], 0, self.nAnt)
            tb.putcol('STATION', [self.siteName,]*self.nAnt, 0, self.nAnt)
            
            for i,ant in enumerate(self.array[0]['ants']):
                #tb.putcell('OFFSET', i, [0.0, 0.0, 0.0])
                tb.putcell('POSITION', i, [ant.x + self.array[0]['center'][0],
                                           ant.y + self.array[0]['center'][1], 
                                           ant.z + self.array[0]['center'][2]])
                #tb.putcell('TYPE', i, 'GROUND-BASED')
                #tb.putcell('DISH_DIAMETER', i, 2.0)
                #tb.putcell('FLAG_ROW', i, False)
                #tb.putcell('MOUNT', i, 'ALT-AZ')
                #tb.putcell('NAME', i, ant.getName())
                #tb.putcell('STATION', i, self.siteName)
                
            tb.flush()
            tb.close()
            
        def _writePolarization(self):
            """
            Write the polarization table.
            """
            
            # Polarization
            
            stks = numpy.array(self.stokes)
            prds = numpy.zeros((2,self.nStokes), dtype=numpy.int32)
            for i,stk in enumerate(self.stokes):
                stks[i] = stk
                if stk > 4:
                    prds[0,i] = ((stk-1) % 4) / 2
                    prds[1,i] = ((stk-1) % 4) % 2
                else:
                    prds[0,i] = 1
                    prds[1,i] = 1
                    
            col1 = tableutil.makearrcoldesc('CORR_TYPE', 0, 1, 
                                            comment='The polarization type for each correlation product, as a Stokes enum.')
            col2 = tableutil.makearrcoldesc('CORR_PRODUCT', 0, 2, 
                                            comment='Indices describing receptors of feed going into correlation')
            col3 = tableutil.makescacoldesc('FLAG_ROW', False, 
                                            comment='flag')
            col4 = tableutil.makescacoldesc('NUM_CORR', self.nStokes, 
                                            comment='Number of correlation products')
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4])
            tb = table("%s/POLARIZATION" % self.basename, desc, nrow=1, ack=False)
            
            tb.putcell('CORR_TYPE', 0, self.stokes)
            tb.putcell('CORR_PRODUCT', 0, prds.T)
            tb.putcell('FLAG_ROW', 0, False)
            tb.putcell('NUM_CORR', 0, self.nStokes)
            
            tb.flush() 
            tb.close()
            
            # Feed
            
            col1  = tableutil.makearrcoldesc('POSITION', 0.0, 1, 
                                             comment='Position of feed relative to feed reference position', 
                                             keywords={'QuantumUnits':['m','m','m'], 
                                                       'MEASINFO':{'type':'position', 'Ref':'ITRF'}
                                                       })
            col2  = tableutil.makearrcoldesc('BEAM_OFFSET', 0.0, 2, 
                                             comment='Beam position offset (on sky but in antennareference frame)', 
                                             keywords={'QuantumUnits':['rad','rad'], 
                                                       'MEASINFO':{'type':'direction', 'Ref':'J2000'}
                                                       })
            col3  = tableutil.makearrcoldesc('POLARIZATION_TYPE', 'X', 1, 
                                             comment='Type of polarization to which a given RECEPTOR responds')
            col4  = tableutil.makearrcoldesc('POL_RESPONSE', 1j, 2,
                                             valuetype='complex',
                                             comment='D-matrix i.e. leakage between two receptors')
            col5  = tableutil.makearrcoldesc('RECEPTOR_ANGLE', 0.0, 1,  
                                             comment='The reference angle for polarization', 
                                             keywords={'QuantumUnits':['rad',]})
            col6  = tableutil.makescacoldesc('ANTENNA_ID', 0, 
                                             comment='ID of antenna in this array')
            col7  = tableutil.makescacoldesc('BEAM_ID', -1, 
                                             comment='Id for BEAM model')
            col8  = tableutil.makescacoldesc('FEED_ID', 0, 
                                             comment='Feed id')
            col9  = tableutil.makescacoldesc('INTERVAL', 0.0, 
                                             comment='Interval for which this set of parameters is accurate', 
                                             keywords={'QuantumUnits':['s',]})
            col10 = tableutil.makescacoldesc('NUM_RECEPTORS', 2, 
                                             comment='Number of receptors on this feed (probably 1 or 2)')
            col11 = tableutil.makescacoldesc('SPECTRAL_WINDOW_ID', -1, 
                                             comment='ID for this spectral window setup')
            col12 = tableutil.makescacoldesc('TIME', 0.0, 
                                             comment='Midpoint of time for which this set of parameters is accurate', 
                                             keywords={'QuantumUnits':['s',], 
                                                       'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                       })
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7, col8, 
                                            col9, col10, col11, col12])
            tb = table("%s/FEED" % self.basename, desc, nrow=self.nAnt, ack=False)
            
            presp = numpy.zeros((self.nAnt,2,2), dtype=numpy.complex64)
            if self.stokes[0] > 8:
                ptype = numpy.tile(['X', 'Y'], (self.nAnt,1))
                presp[:,0,0] = 1.0
                presp[:,0,1] = 0.0
                presp[:,1,0] = 0.0
                presp[:,1,1] = 1.0
            elif self.stokes[0] > 4:
                ptype = numpy.tile(['R', 'L'], (self.nAnt,1))
                presp[:,0,0] = 1.0
                presp[:,0,1] = -1.0j
                presp[:,1,0] = 1.0j
                presp[:,1,1] = 1.0
            else:
                ptype = numpy.tile(['X', 'Y'], (self.nAnt,1))
                presp[:,0,0] = 1.0
                presp[:,0,1] = 0.0
                presp[:,1,0] = 0.0
                presp[:,1,1] = 1.0
                
            tb.putcol('POSITION', numpy.zeros((self.nAnt,3)), 0, self.nAnt)
            tb.putcol('BEAM_OFFSET', numpy.zeros((self.nAnt,2,2)), 0, self.nAnt)
            tb.putcol('POLARIZATION_TYPE', ptype, 0, self.nAnt)
            tb.putcol('POL_RESPONSE', presp, 0, self.nAnt)
            tb.putcol('RECEPTOR_ANGLE', numpy.zeros((self.nAnt,2)), 0, self.nAnt)
            tb.putcol('ANTENNA_ID', list(range(self.nAnt)), 0, self.nAnt)
            tb.putcol('BEAM_ID', [-1,]*self.nAnt, 0, self.nAnt)
            tb.putcol('FEED_ID', [0,]*self.nAnt, 0, self.nAnt)
            tb.putcol('INTERVAL', [0.0,]*self.nAnt, 0, self.nAnt)
            tb.putcol('NUM_RECEPTORS', [2,]*self.nAnt, 0, self.nAnt)
            tb.putcol('SPECTRAL_WINDOW_ID', [-1,]*self.nAnt, 0, self.nAnt)
            tb.putcol('TIME', [0.0,]*self.nAnt, 0, self.nAnt)
            
            tb.flush() 
            tb.close()
            
        def _writeObservation(self):
            """
            Write the observation table.
            """
            
            # Observation
            
            col1 = tableutil.makearrcoldesc('TIME_RANGE', 0.0, 1, 
                                            comment='Start and end of observation', 
                                            keywords={'QuantumUnits':['s',], 
                                                      'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                      })
            col2 = tableutil.makearrcoldesc('LOG', 'none', 1, 
                                            comment='Observing log')
            col3 = tableutil.makearrcoldesc('SCHEDULE', 'none', 1, 
                                            comment='Observing schedule')
            col4 = tableutil.makescacoldesc('FLAG_ROW', False, 
                                            comment='Row flag')
            col5 = tableutil.makescacoldesc('OBSERVER', 'ZASKY', 
                                            comment='Name of observer(s)')
            col6 = tableutil.makescacoldesc('PROJECT', 'ZASKY', 
                                            comment='Project identification string')
            col7 = tableutil.makescacoldesc('RELEASE_DATE', 0.0, 
                                            comment='Release date when data becomes public', 
                                            keywords={'QuantumUnits':['s',], 
                                                      'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                      })
            col8 = tableutil.makescacoldesc('SCHEDULE_TYPE', 'none', 
                                            comment='Observing schedule type')
            col9 = tableutil.makescacoldesc('TELESCOPE_NAME', self.siteName, 
                                            comment='Telescope Name (e.g. WSRT, VLBA)')
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7, col8, col9])
            tb = table("%s/OBSERVATION" % self.basename, desc, nrow=1, ack=False)
            
            tStart = astro.taimjd_to_utcjd(self.data[ 0].obsTime) - astro.MJD_OFFSET
            tStop  = astro.taimjd_to_utcjd(self.data[-1].obsTime) - astro.MJD_OFFSET
            
            tb.putcell('TIME_RANGE', 0, [tStart*86400, tStop*86400])
            tb.putcell('LOG', 0, 'Not provided')
            tb.putcell('SCHEDULE', 0, 'Not provided')
            tb.putcell('FLAG_ROW', 0, False)
            tb.putcell('OBSERVER', 0, 'ZASKY')
            tb.putcell('PROJECT', 0, 'ZASKY')
            tb.putcell('RELEASE_DATE', 0, tStop*86400)
            tb.putcell('SCHEDULE_TYPE', 0, 'None')
            tb.putcell('TELESCOPE_NAME', 0, self.siteName)
            
            tb.flush() 
            tb.close()
            
            # Source
            
            arrayGeo = astro.rect_posn(*self.array[0]['center'])
            arrayGeo = astro.get_geo_from_rect(arrayGeo)
            
            obs = ephem.Observer()
            obs.lat = arrayGeo.lat * numpy.pi/180
            obs.lon = arrayGeo.lng * numpy.pi/180
            obs.elev = arrayGeo.elv * numpy.pi/180
            obs.pressure = 0
            
            nameList = []
            posList = []
            sourceID = 0
            for dataSet in self.data:
                if dataSet.pol == self.stokes[0]:
                    utc = astro.taimjd_to_utcjd(dataSet.obsTime)
                    date = astro.get_date(utc)
                    date.hours = 0
                    date.minutes = 0
                    date.seconds = 0
                    utc0 = date.to_jd()
                    
                    obs.date = utc - astro.DJD_OFFSET
                    
                    try:
                        currSourceName = dataSet.source.name
                    except AttributeError:
                        currSourceName = dataSet.source
                    
                    if currSourceName not in nameList:
                        sourceID += 1
                        
                        if dataSet.source == 'z':
                            ## Zenith pointings
                            equ = astro.equ_posn( obs.sidereal_time()*180/numpy.pi, obs.lat*180/numpy.pi )
                            
                            # format 'source' name based on local sidereal time
                            raHms = astro.deg_to_hms(equ.ra)
                            (tsecs, secs) = math.modf(raHms.seconds)
                            
                            name = "ZA%02d%02d%02d%01d" % (raHms.hours, raHms.minutes, int(secs), int(tsecs * 10.0))
                            equPo = astro.get_equ_prec2(equ, utc, astro.J2000_UTC_JD)
                            
                        else:
                            ## Real-live sources (ephem.Body instances)
                            name = dataSet.source.name
                            equ = astro.equ_posn(dataSet.source.ra*180/numpy.pi, dataSet.source.dec*180/numpy.pi)
                            equPo = astro.equ_posn(dataSet.source.a_ra*180/numpy.pi, dataSet.source.a_dec*180/numpy.pi)
                            
                        # J2000 zenith equatorial coordinates
                        posList.append( [equPo.ra*numpy.pi/180, equPo.dec*numpy.pi/180] )
                        
                        # name
                        nameList.append(name)
                        
            nSource = len(nameList)
            
            # Save these for later since we might need them
            self._sourceTable = nameList
            
            col1  = tableutil.makearrcoldesc('DIRECTION', 0.0, 1, 
                                             comment='Direction (e.g. RA, DEC).', 
                                             keywords={'QuantumUnits':['rad','rad'], 
                                                       'MEASINFO':{'type':'direction', 'Ref':'J2000'}
                                                       })
            col2  = tableutil.makearrcoldesc('PROPER_MOTION', 0.0, 1, 
                                             comment='Proper motion', 
                                             keywords={'QuantumUnits':['rad/s',]})
            col3  = tableutil.makescacoldesc('CALIBRATION_GROUP', 0, 
                                             comment='Number of grouping for calibration purpose.')
            col4  = tableutil.makescacoldesc('CODE', "none", 
                                             comment='Special characteristics of source, e.g. Bandpass calibrator')
            col5  = tableutil.makescacoldesc('INTERVAL', 0.0, 
                                             comment='Interval of time for which this set of parameters is accurate', 
                                             keywords={'QuantumUnits':['s',]})
            col6  = tableutil.makescacoldesc('NAME', "none", 
                                             comment='Name of source as given during observations')
            col7  = tableutil.makescacoldesc('NUM_LINES', 0, 
                                             comment='Number of spectral lines')
            col8  = tableutil.makescacoldesc('SOURCE_ID', 0, 
                                             comment='Source id')
            col9  = tableutil.makescacoldesc('SPECTRAL_WINDOW_ID', -1, 
                                             comment='ID for this spectral window setup')
            col10 = tableutil.makescacoldesc('TIME', 0.0,
                                             comment='Midpoint of time for which this set of parameters is accurate.', 
                                             keywords={'QuantumUnits':['s',], 
                                                       'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                       })
            col11 = tableutil.makearrcoldesc('TRANSITION', 'none', 1, 
                                             comment='Line Transition name')
            col12 = tableutil.makearrcoldesc('REST_FREQUENCY', 1.0, 1, 
                                             comment='Line rest frequency', 
                                             keywords={'QuantumUnits':['Hz',], 
                                                       'MEASINFO':{'type':'frequency', 
                                                                   'Ref':'LSRK'}
                                                       })
            col13 = tableutil.makearrcoldesc('SYSVEL', 1.0, 1, 
                                             comment='Systemic velocity at reference', 
                                             keywords={'QuantumUnits':['m/s',], 
                                                       'MEASINFO':{'type':'radialvelocity', 
                                                                   'Ref':'LSRK'}
                                                       })
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7, col8, col9, 
                                          col10, col11, col12, col13])
            tb = table("%s/SOURCE" % self.basename, desc, nrow=nSource, ack=False)
            
            for i in xrange(nSource):
                tb.putcell('DIRECTION', i, posList[i])
                tb.putcell('PROPER_MOTION', i, [0.0, 0.0])
                tb.putcell('CALIBRATION_GROUP', i, 0)
                tb.putcell('CODE', i, 'none')
                tb.putcell('INTERVAL', i, 0.0)
                tb.putcell('NAME', i, nameList[i])
                tb.putcell('NUM_LINES', i, 0)
                tb.putcell('SOURCE_ID', i, i)
                tb.putcell('SPECTRAL_WINDOW_ID', i, -1)
                tb.putcell('TIME', i, (tStart+tStop)/2*86400)
                #tb.putcell('TRANSITION', i, [])
                #tb.putcell('REST_FREQUENCY', i, [])
                #tb.putcell('SYSVEL', i, [])
                
            tb.flush()
            tb.close()
            
            # Field
            
            col1 = tableutil.makearrcoldesc('DELAY_DIR', 0.0, 2, 
                                            comment='Direction of delay center (e.g. RA, DEC)as polynomial in time.', 
                                            keywords={'QuantumUnits':['rad','rad'], 
                                                      'MEASINFO':{'type':'direction', 'Ref':'J2000'}
                                                      })
            col2 = tableutil.makearrcoldesc('PHASE_DIR', 0.0, 2, 
                                            comment='Direction of phase center (e.g. RA, DEC).', 
                                            keywords={'QuantumUnits':['rad','rad'], 
                                                      'MEASINFO':{'type':'direction', 'Ref':'J2000'}
                                                      })
            col3 = tableutil.makearrcoldesc('REFERENCE_DIR', 0.0, 2, 
                                            comment='Direction of REFERENCE center (e.g. RA, DEC).as polynomial in time.', 
                                            keywords={'QuantumUnits':['rad','rad'], 
                                                      'MEASINFO':{'type':'direction', 'Ref':'J2000'}
                                                      })
            col4 = tableutil.makescacoldesc('CODE', "none", 
                                            comment='Special characteristics of field, e.g. Bandpass calibrator')
            col5 = tableutil.makescacoldesc('FLAG_ROW', False, 
                                            comment='Row Flag')
            col6 = tableutil.makescacoldesc('NAME', "none", 
                                            comment='Name of this field')
            col7 = tableutil.makescacoldesc('NUM_POLY', 0, 
                                            comment='Polynomial order of _DIR columns')
            col8 = tableutil.makescacoldesc('SOURCE_ID', 0, 
                                            comment='Source id')
            col9 = tableutil.makescacoldesc('TIME', (tStart+tStop)/2, 
                                            comment='Time origin for direction and rate', 
                                            keywords={'QuantumUnits':['s',],
                                                      'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                      })
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7, col8, col9])
            tb = table("%s/FIELD" % self.basename, desc, nrow=nSource, ack=False)
            
            for i in xrange(nSource):
                tb.putcell('DELAY_DIR', i, numpy.array([posList[i],]))
                tb.putcell('PHASE_DIR', i, numpy.array([posList[i],]))
                tb.putcell('REFERENCE_DIR', i, numpy.array([posList[i],]))
                tb.putcell('CODE', i, 'None')
                tb.putcell('FLAG_ROW', i, False)
                tb.putcell('NAME', i, nameList[i])
                tb.putcell('NUM_POLY', i, 0)
                tb.putcell('SOURCE_ID', i, i)
                tb.putcell('TIME', i, (tStart+tStop)/2*86400)
                
            tb.flush()
            tb.close()
            
        def _writeSpectralWindow(self):
            """
            Write the spectral window table.
            """
            
            # Spectral Window
            
            nBand = len(self.freq)
            
            col1  = tableutil.makescacoldesc('MEAS_FREQ_REF', 0, 
                                             comment='Frequency Measure reference')
            col2  = tableutil.makearrcoldesc('CHAN_FREQ', 0.0, 1, 
                                             comment='Center frequencies for each channel in the data matrix', 
                                             keywords={'QuantumUnits':['Hz',], 
                                                       'MEASINFO':{'type':'frequency', 
                                                                   'VarRefCol':'MEAS_FREQ_REF', 
                                                                   'TabRefTypes':['REST','LSRK','LSRD','BARY','GEO','TOPO','GALACTO','LGROUP','CMB','Undefined'],
                                                                   'TabRefCodes':[0,1,2,3,4,5,6,7,8,64]}
                                                       })
            col3  = tableutil.makescacoldesc('REF_FREQUENCY', self.refVal, 
                                             comment='The reference frequency', 
                                             keywords={'QuantumUnits':['Hz',], 
                                                       'MEASINFO':{'type':'frequency', 
                                                                   'VarRefCol':'MEAS_FREQ_REF', 
                                                                   'TabRefTypes':['REST','LSRK','LSRD','BARY','GEO','TOPO','GALACTO','LGROUP','CMB','Undefined'],
                                                                   'TabRefCodes':[0,1,2,3,4,5,6,7,8,64]}
                                                       })
            col4  = tableutil.makearrcoldesc('CHAN_WIDTH', 0.0, 1, 
                                             comment='Channel width for each channel', 
                                             keywords={'QuantumUnits':['Hz',]})
            col5  = tableutil.makearrcoldesc('EFFECTIVE_BW', 0.0, 1, 
                                             comment='Effective noise bandwidth of each channel', 
                                             keywords={'QuantumUnits':['Hz',]})
            col6  = tableutil.makearrcoldesc('RESOLUTION', 0.0, 1, 
                                             comment='The effective noise bandwidth for each channel', 
                                             keywords={'QuantumUnits':['Hz',]})
            col7  = tableutil.makescacoldesc('FLAG_ROW', False, 
                                             comment='flag')
            col8  = tableutil.makescacoldesc('FREQ_GROUP', 1, 
                                             comment='Frequency group')
            col9  = tableutil.makescacoldesc('FREQ_GROUP_NAME', "group1", 
                                             comment='Frequency group name')
            col10 = tableutil.makescacoldesc('IF_CONV_CHAIN', 0, 
                                             comment='The IF conversion chain number')
            col11 = tableutil.makescacoldesc('NAME', "%i channels" % self.nChan, 
                                             comment='Spectral window name')
            col12 = tableutil.makescacoldesc('NET_SIDEBAND', 0, 
                                             comment='Net sideband')
            col13 = tableutil.makescacoldesc('NUM_CHAN', 0, 
                                             comment='Number of spectral channels')
            col14 = tableutil.makescacoldesc('TOTAL_BANDWIDTH', 0.0, 
                                             comment='The total bandwidth for this window', 
                                             keywords={'QuantumUnits':['Hz',]})
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7, col8, col9, 
                                            col10, col11, col12, col13, col14])
            tb = table("%s/SPECTRAL_WINDOW" % self.basename, desc, nrow=nBand, ack=False)
            
            for i,freq in enumerate(self.freq):
                tb.putcell('MEAS_FREQ_REF', i, 0)
                tb.putcell('CHAN_FREQ', i, self.refVal + freq.bandFreq + numpy.arange(self.nChan)*self.channelWidth)
                tb.putcell('REF_FREQUENCY', i, self.refVal)
                tb.putcell('CHAN_WIDTH', i, [freq.chWidth for j in xrange(self.nChan)])
                tb.putcell('EFFECTIVE_BW', i, [freq.chWidth for j in xrange(self.nChan)])
                tb.putcell('RESOLUTION', i, [freq.chWidth for j in xrange(self.nChan)])
                tb.putcell('FLAG_ROW', i, False)
                tb.putcell('FREQ_GROUP', i, i+1)
                tb.putcell('FREQ_GROUP_NAME', i, 'group%i' % (i+1))
                tb.putcell('IF_CONV_CHAIN', i, i)
                tb.putcell('NAME', i, "IF %i, %i channels" % (i+1, self.nChan))
                tb.putcell('NET_SIDEBAND', i, 0)
                tb.putcell('NUM_CHAN', i, self.nChan)
                tb.putcell('TOTAL_BANDWIDTH', i, freq.totalBW)
                
            tb.flush()
            tb.close()
            
        def _writeMain(self):
            """
            Write the main table.
            """
            
            # Main
            
            nBand = len(self.freq)
            
            arrayGeo = astro.rect_posn(*self.array[0]['center'])
            arrayGeo = astro.get_geo_from_rect(arrayGeo)
            
            obs = ephem.Observer()
            obs.lat = arrayGeo.lat * numpy.pi/180
            obs.lon = arrayGeo.lng * numpy.pi/180
            obs.elev = arrayGeo.elv * numpy.pi/180
            obs.pressure = 0
            
            mapper = self.array[0]['mapper']
            
            col1  = tableutil.makearrcoldesc('UVW', 0.0, 1, 
                                             comment='Vector with uvw coordinates (in meters)', 
                                             keywords={'QuantumUnits':['m','m','m'], 
                                                       'MEASINFO':{'type':'uvw', 'Ref':'ITRF'}
                                                       })
            col2  = tableutil.makearrcoldesc('FLAG', False, 2, 
                                             comment='The data flags, array of bools with same shape as data')
            col3  = tableutil.makearrcoldesc('FLAG_CATEGORY', False, 3,  
                                             comment='The flag category, NUM_CAT flags for each datum', 
                                             keywords={'CATEGORY':['',]})
            col4  = tableutil.makearrcoldesc('WEIGHT', 1.0, 1, 
                                             valuetype='float', 
                                             comment='Weight for each polarization spectrum')
            col5  = tableutil.makearrcoldesc('SIGMA', 9999., 1, 
                                             valuetype='float', 
                                             comment='Estimated rms noise for channel with unity bandpass response')
            col6  = tableutil.makescacoldesc('ANTENNA1', 0, 
                                             comment='ID of first antenna in interferometer')
            col7  = tableutil.makescacoldesc('ANTENNA2', 0, 
                                             comment='ID of second antenna in interferometer')
            col8  = tableutil.makescacoldesc('ARRAY_ID', 0, 
                                             comment='ID of array or subarray')
            col9  = tableutil.makescacoldesc('DATA_DESC_ID', 0, 
                                             comment='The data description table index')
            col10 = tableutil.makescacoldesc('EXPOSURE', 0.0, 
                                             comment='he effective integration time', 
                                             keywords={'QuantumUnits':['s',]})
            col11 = tableutil.makescacoldesc('FEED1', 0, 
                                             comment='The feed index for ANTENNA1')
            col12 = tableutil.makescacoldesc('FEED2', 0, 
                                             comment='The feed index for ANTENNA2')
            col13 = tableutil.makescacoldesc('FIELD_ID', 0, 
                                             comment='Unique id for this pointing')
            col14 = tableutil.makescacoldesc('FLAG_ROW', False, 
                                             comment='Row flag - flag all data in this row if True')
            col15 = tableutil.makescacoldesc('INTERVAL', 0.0, 
                                             comment='The sampling interval', 
                                             keywords={'QuantumUnits':['s',]})
            col16 = tableutil.makescacoldesc('OBSERVATION_ID', 0, 
                                             comment='ID for this observation, index in OBSERVATION table')
            col17 = tableutil.makescacoldesc('PROCESSOR_ID', -1, 
                                             comment='Id for backend processor, index in PROCESSOR table')
            col18 = tableutil.makescacoldesc('SCAN_NUMBER', 1, 
                                             comment='Sequential scan number from on-line system')
            col19 = tableutil.makescacoldesc('STATE_ID', -1, 
                                             comment='ID for this observing state')
            col20 = tableutil.makescacoldesc('TIME', 0.0, 
                                             comment='Modified Julian Day', 
                                             keywords={'QuantumUnits':['s',],
                                                       'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                       })
            col21 = tableutil.makescacoldesc('TIME_CENTROID', 0.0, 
                                             comment='Modified Julian Day', 
                                             keywords={'QuantumUnits':['s',],
                                                       'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                       })
            col22 = tableutil.makearrcoldesc("DATA", 0j, 2, 
                                             valuetype='complex',
                                             comment='The data column')
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7, col8, col9, 
                                            col10, col11, col12, col13, col14, col15, col16, 
                                            col17, col18, col19, col20, col21, col22])
            tb = table("%s" % self.basename, desc, nrow=0, ack=False)
            
            i = 0
            s = 1
            _sourceTable = []
            for dataSet in self.data:
                # Sort the data by packed baseline
                try:
                    order
                except NameError:
                    order = dataSet.argsort(mapper=mapper, shift=16)
                    
                # Deal with defininig the values of the new data set
                if dataSet.pol == self.stokes[0]:
                    ## Figure out the new date/time for the observation
                    utc = astro.taimjd_to_utcjd(dataSet.obsTime)
                    date = astro.get_date(utc)
                    date.hours = 0
                    date.minutes = 0
                    date.seconds = 0
                    utc0 = date.to_jd()
                    
                    ## Update the observer so we can figure out where the source is
                    obs.date = utc - astro.DJD_OFFSET
                    if dataSet.source == 'z':
                        ### Zenith pointings
                        equ = astro.equ_posn( obs.sidereal_time()*180/numpy.pi, obs.lat*180/numpy.pi )
                        
                        ### format 'source' name based on local sidereal time
                        raHms = astro.deg_to_hms(equ.ra)
                        (tsecs, secs) = math.modf(raHms.seconds)
                        name = "ZA%02d%02d%02d%01d" % (raHms.hours, raHms.minutes, int(secs), int(tsecs * 10.0))
                    else:
                        ### Real-live sources (ephem.Body instances)
                        name = dataSet.source.name
                        
                    ## Update the source ID
                    try:
                        sourceID = _sourceTable.index(name)
                    except ValueError:
                        _sourceTable.append(name)
                        sourceID = _sourceTable.index(name)
                        
                    ## Compute the uvw coordinates of all baselines
                    if dataSet.source == 'z':
                        HA = 0.0
                        dec = equ.dec
                    else:
                        HA = (obs.sidereal_time() - dataSet.source.ra) * 12/numpy.pi
                        dec = dataSet.source.dec * 180/numpy.pi
                    uvwCoords = dataSet.getUVW(HA, dec, obs)
                    
                    ## Populate the metadata
                    ### Add in the baselines
                    try:
                        ant1List
                        ant2List
                    except NameError:
                        a1List, a2List = [], []
                        for o in order:
                            antenna1, antenna2 = dataSet.baselines[o]
                            a1 = mapper.index(antenna1.stand.id)
                            a2 = mapper.index(antenna2.stand.id)
                            a1List.append( a1 )
                            a2List.append( a2 )
                        ant1List = a1List
                        ant2List = a2List
                        
                    ### Add in the new u, v, and w coordinates
                    uvwList = uvwCoords[order,:]
                    
                    ### Add in the new date/time and integration time
                    #timeList = [utc - astro.MJD_OFFSET for bl in dataSet.baselines]
                    intTimeList = [dataSet.intTime for bl in dataSet.baselines]
                    timeList = [(utc - astro.MJD_OFFSET)*86400 + dataSet.intTime/2.0 for bl in dataSet.baselines]
                    
                    ### Add in the new new source ID and name
                    sourceList = [sourceID for bl in dataSet.baselines]
                     
                    ### Zero out the visibility data
                    try:
                        matrix.shape = (len(order), self.nStokes, nBand*self.nChan)
                        matrix *= 0.0
                    except NameError:
                        matrix = numpy.zeros((len(order), self.nStokes, self.nChan*nBand), dtype=numpy.complex64)
                        
                # Save the visibility data in the right order
                matrix[:,self.stokes.index(dataSet.pol),:] = dataSet.visibilities[order,:]
                
                # Deal with saving the data once all of the polarizations have been added to 'matrix'
                if dataSet.pol == self.stokes[-1]:
                    nBL = uvwList.shape[0]
                    tb.addrows(nBand*nBL)
                    
                    matrix.shape = (len(order), self.nStokes, nBand, self.nChan)
                    
                    for j in xrange(nBand):
                        fg = numpy.zeros((nBL,self.nStokes,self.nChan), dtype=numpy.bool)
                        fc = numpy.zeros((nBL,self.nStokes,self.nChan,1), dtype=numpy.bool)
                        wg = numpy.ones((nBL,self.nStokes))
                        sg = numpy.ones((nBL,self.nStokes))*9999
                        
                        tb.putcol('UVW', uvwList, i, nBL)
                        tb.putcol('FLAG', fg.transpose(0,2,1), i, nBL)
                        tb.putcol('FLAG_CATEGORY', fc.transpose(0,3,2,1), i, nBL)
                        tb.putcol('WEIGHT', wg, i, nBL)
                        tb.putcol('SIGMA', sg, i, nBL)
                        tb.putcol('ANTENNA1', ant1List, i, nBL)
                        tb.putcol('ANTENNA2', ant2List, i, nBL)
                        tb.putcol('ARRAY_ID', [0,]*nBL, i, nBL)
                        tb.putcol('DATA_DESC_ID', [j,]*nBL, i, nBL)
                        tb.putcol('EXPOSURE', intTimeList, i, nBL)
                        tb.putcol('FEED1', [0,]*nBL, i, nBL)
                        tb.putcol('FEED2', [0,]*nBL, i, nBL)
                        tb.putcol('FIELD_ID', sourceList, i, nBL)
                        tb.putcol('FLAG_ROW', [False,]*nBL, i, nBL)
                        tb.putcol('INTERVAL', intTimeList, i, nBL)
                        tb.putcol('OBSERVATION_ID', [0,]*nBL, i, nBL)
                        tb.putcol('PROCESSOR_ID', [-1,]*nBL, i, nBL)
                        tb.putcol('SCAN_NUMBER', [s,]*nBL, i, nBL)
                        tb.putcol('STATE_ID', [-1,]*nBL, i, nBL)
                        tb.putcol('TIME', timeList, i, nBL)
                        tb.putcol('TIME_CENTROID', timeList, i, nBL)
                        tb.putcol('DATA', matrix[...,j,:].transpose(0,2,1), i, nBL)
                        i += nBL
                    s += 1
                    
            tb.flush()
            tb.close()
            
            # Data description
            
            col1 = tableutil.makescacoldesc('FLAG_ROW', False, 
                                            comment='Flag this row')
            col2 = tableutil.makescacoldesc('POLARIZATION_ID', 0, 
                                            comment='Pointer to polarization table')
            col3 = tableutil.makescacoldesc('SPECTRAL_WINDOW_ID', 0, 
                                            comment='Pointer to spectralwindow table')
            
            desc = tableutil.maketabdesc([col1, col2, col3])
            tb = table("%s/DATA_DESCRIPTION" % self.basename, desc, nrow=nBand, ack=False)
            
            for i in xrange(nBand):
                tb.putcell('FLAG_ROW', i, False)
                tb.putcell('POLARIZATION_ID', i, 0)
                tb.putcell('SPECTRAL_WINDOW_ID', i, i)
                
            tb.flush()
            tb.close()
            
        def _writeMisc(self):
            """
            Write the other tables that are part of the measurement set but 
            don't contain anything by default.
            """
            
            # Flag command
            
            col1 = tableutil.makescacoldesc('TIME', 0.0, 
                                            comment='Midpoint of interval for which this flag is valid', 
                                            keywords={'QuantumUnits':['s',], 
                                                      'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                      })
            col2 = tableutil.makescacoldesc('INTERVAL', 0.0, 
                                            comment='Time interval for which this flag is valid', 
                                            keywords={'QuantumUnits':['s',]})
            col3 = tableutil.makescacoldesc('TYPE', 'flag', 
                                            comment='Type of flag (FLAG or UNFLAG)')
            col4 = tableutil.makescacoldesc('REASON', 'reason', 
                                            comment='Flag reason')
            col5 = tableutil.makescacoldesc('LEVEL', 0, 
                                            comment='Flag level - revision level')
            col6 = tableutil.makescacoldesc('SEVERITY', 0, 
                                            comment='Severity code (0-10)')
            col7 = tableutil.makescacoldesc('APPLIED', False, 
                                            comment='True if flag has been applied to main table')
            col8 = tableutil.makescacoldesc('COMMAND', 'command', 
                                            comment='Flagging command')
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7, col8])
            tb = table("%s/FLAG_CMD" % self.basename, desc, nrow=0, ack=False)
            
            tb.flush()
            tb.close()
            
            # History
            
            col1 = tableutil.makescacoldesc('TIME', 0.0, 
                                            comment='Timestamp of message', 
                                            keywords={'QuantumUnits':['s',], 
                                                      'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                      })
            col2 = tableutil.makescacoldesc('OBSERVATION_ID', 0, 
                                            comment='Observation id (index in OBSERVATION table)')
            col3 = tableutil.makescacoldesc('MESSAGE', 'message', 
                                            comment='Log message')
            col4 = tableutil.makescacoldesc('PRIORITY', 'NORMAL', 
                                            comment='Message priority')
            col5 = tableutil.makescacoldesc('ORIGIN', 'origin', 
                                            comment='(Source code) origin from which message originated')
            col6 = tableutil.makescacoldesc('OBJECT_ID', 0, 
                                            comment='Originating ObjectID')
            col7 = tableutil.makescacoldesc('APPLICATION', 'application', 
                                            comment='Application name')
            col8 = tableutil.makearrcoldesc('CLI_COMMAND', 'command', 1, 
                                            comment='CLI command sequence')
            col9 = tableutil.makearrcoldesc('APP_PARAMS', 'params', 1, 
                                            comment='Application parameters')
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7, col8, col9])
            tb = table("%s/HISTORY" % self.basename, desc, nrow=0, ack=False)
            
            tb.flush()
            tb.close()
            
            # POINTING
            
            col1 = tableutil.makescacoldesc('ANTENNA_ID', 0, 
                                            comment='Antenna Id')
            col2 = tableutil.makescacoldesc('TIME', 0.0, 
                                            comment='Time interval midpoint', 
                                            keywords={'QuantumUnits':['s',], 
                                                      'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                      })
            col3 = tableutil.makescacoldesc('INTERVAL', 0.0, 
                                            comment='Time interval', 
                                            keywords={'QuantumUnits':['s',]})
            col4 = tableutil.makescacoldesc('NAME', 'name', 
                                            comment='Pointing position name')
            col5 = tableutil.makescacoldesc('NUM_POLY', 0, 
                                            comment='Series order')
            col6 = tableutil.makescacoldesc('TIME_ORIGIN', 0.0, 
                                            comment='Time origin for direction', 
                                            keywords={'QuantumUnits':['s',], 
                                                      'MEASINFO':{'type':'epoch', 'Ref':'UTC'}
                                                      })
            col7 = tableutil.makearrcoldesc('DIRECTION', 0.0, 2, 
                                            comment='Antenna pointing direction as polynomial in time', 
                                            keywords={'QuantumUnits':['rad','rad'], 
                                                      'MEASINFO':{'type':'direction', 'Ref':'J2000'}
                                                      })
            col8 = tableutil.makearrcoldesc('TARGET', 0.0, 2, 
                                            comment='target direction as polynomial in time',
                                            keywords={'QuantumUnits':['rad','rad'], 
                                                      'MEASINFO':{'type':'direction', 'Ref':'J2000'}
                                                      })
            col9 = tableutil.makescacoldesc('TRACKING', True, 
                                            comment='Tracking flag - True if on position')
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7, col8, col9])
            tb = table("%s/POINTING" % self.basename, desc, nrow=0, ack=False)
            
            tb.flush()
            tb.close()
            
            # Processor
            
            col1 = tableutil.makescacoldesc('TYPE', 'type', 
                                            comment='Processor type')
            col2 = tableutil.makescacoldesc('SUB_TYPE', 'subtype', 
                                            comment='Processor sub type')
            col3 = tableutil.makescacoldesc('TYPE_ID', 0, 
                                            comment='Processor type id')
            col4 = tableutil.makescacoldesc('MODE_ID', 0, 
                                            comment='Processor mode id')
            col5 = tableutil.makescacoldesc('FLAG_ROW', False, 
                                            comment='flag')
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5])
            tb = table("%s/PROCESSOR" % self.basename, desc, nrow=0, ack=False)
            
            tb.flush()
            tb.close()
            
            # State
            
            col1 = tableutil.makescacoldesc('SIG', True, 
                                            comment='True for a source observation')
            col2 = tableutil.makescacoldesc('REF', False, 
                                            comment='True for a reference observation')
            col3 = tableutil.makescacoldesc('CAL', 0.0, 
                                            comment='Noise calibration temperature', 
                                            keywords={'QuantumUnits':['K',]})
            col4 = tableutil.makescacoldesc('LOAD', 0.0, 
                                            comment='Load temperature', 
                                            keywords={'QuantumUnits':['K',]})
            col5 = tableutil.makescacoldesc('SUB_SCAN', 0, 
                                            comment='Sub scan number, relative to scan number')
            col6 = tableutil.makescacoldesc('OBS_MODE', 'mode', 
                                            comment='Observing mode, e.g., OFF_SPECTRUM')
            col7 = tableutil.makescacoldesc('FLAG_ROW', False, 
                                            comment='Row flag')
            
            desc = tableutil.maketabdesc([col1, col2, col3, col4, col5, col6, col7])
            tb = table("%s/STATE" % self.basename, desc, nrow=0, ack=False)
            
            tb.flush()
            tb.close()
            
except ImportError:
    import warnings
    warnings.warn('Cannot import casacore.tables, MS support disabled', ImportWarning)
    
    class MS(object):
        """
        Class for storing visibility data and writing the data, along with array
        geometry, frequency setup, etc., to a CASA measurement set.
        """
        
        def __init__(self, filename, refTime=0.0, verbose=False, memmap=None, clobber=False):
            """
            Initialize a new FITS IDI object using a filename and a reference time 
            given in seconds since the UNIX 1970 ephem, a python datetime object, or a 
            string in the format of 'YYYY-MM-DDTHH:MM:SS'.
            
            .. versionchanged:: 1.1.2
                Added the 'memmap' and 'clobber' keywords to control if the file
                is memory mapped and whether or not to overwrite an existing file, 
                respectively.
            """
            
            raise RuntimeError("Cannot import casacore.tables, MS support disabled")



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""Application to calculate real-time ephemeris for a LWA site."""

import math
import time
import curses
import argparse

from lsl import astro
from lsl import transform
from lsl.common import stations

__revision__  = "$Revision: 94 $"
__version__   = "dev"
__author__    = "P.S.Ray"


def restorescreen():
    curses.nocbreak()
    curses.echo()
    curses.endwin()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('-s', '--site', type=str, default='LWA-1',
                        help='site name')
    args = parser.parse_args()
    
    # setup transform objects
    site = args.site.lower().replace('-', '')
    if site == 'lwa1':
        station = stations.lwa1
    elif site == 'lwasv':
        station = stations.lwasv
    elif site == 'ovrolwa':
        station = stations.lwa1
        station.name = 'OVRO-LWA'
        station.lat, station.lon, station.elev = ('37.2397808', '-118.2816819', 1183.4839)
    else:
        raise RuntimeError("Unknown site name: %s" % site)
    site = transform.GeographicalPosition((station.long*180.0/math.pi, station.lat*180.0/math.pi), name=station.name)
    
    sun_pos = transform.PlanetaryPosition('Sun')
    sun_dir = transform.PointingDirection(sun_pos, site)
    
    jup_pos = transform.PlanetaryPosition('Jupiter')
    jup_dir = transform.PointingDirection(jup_pos, site)
    
    gal_pos = transform.CelestialPosition((astro.hms(17, 42, 0.1),
    astro.dms(True, 28, 55, 8)), name = 'SgrA')
    gal_dir = transform.PointingDirection(gal_pos, site)
    
    tau_pos = transform.CelestialPosition((astro.hms(5, 34, 0.5),
    astro.dms(False, 22, 0, 52)), name = 'TauA')
    tau_dir = transform.PointingDirection(tau_pos, site)
    
    cyg_pos = transform.CelestialPosition((astro.hms(19, 59, 0.8),
    astro.dms(False, 40, 44, 2)), name = 'CygA')
    cyg_dir = transform.PointingDirection(cyg_pos, site)
    
    cas_pos = transform.CelestialPosition((astro.hms(23, 23, 0.7),
    astro.dms(False, 58, 49, 16)), name = 'CasA')
    cas_dir = transform.PointingDirection(cas_pos, site)
    
    try:
        # setup display screen
        screen = curses.initscr()
        curses.start_color()
        curses.noecho()
        curses.cbreak()
        screen.nodelay(1)
        screen.border()
        
        if curses.has_colors():
            curses. use_default_colors()
            
            curses.init_pair(1, -1, -1)
            curses.init_pair(2, curses.COLOR_RED, -1)
            curses.init_pair(3, curses.COLOR_GREEN, -1)
            curses.init_pair(4, curses.COLOR_BLUE, -1)
            curses.init_pair(5, curses.COLOR_YELLOW, -1)
            curses.init_pair(6, curses.COLOR_CYAN, -1)
            
            std = curses.color_pair(1)
            dwn = curses.color_pair(2)
            up  = curses.color_pair(3)
            st  = curses.color_pair(4) | curses.A_BOLD
            rs  = curses.color_pair(5) | curses.A_BOLD
            inf = curses.color_pair(6)
        else:
            std = curses.A_NORMAL
            dwn = curses.A_NORMAL
            up  = curses.A_NORMAL
            st  = curses.A_NORMAL
            rs  = curses.A_NORMAL
            inf = curses.A_NORMAL
            
        strdict = {}
        while (True):
            screen.clear()
            
            # Build string to print
            # update current time and local sidereal time
            currentTime = transform.Time.from_system() 
                        
            strdict['utc'] = "%-25s" % currentTime.utc_str
            strdict['mjdutc'] = "%.5f" % currentTime.utc_mjd
            strdict['loc'] = site.name
            (lng, lat) = site.geo.format()
            strdict['lat'] = "%-15s" % lat
            strdict['long'] = "%-15s" % lng
            strdict['lst'] = "%-15s" % astro.deg_to_dms(site.sidereal(currentTime))
            
            # calculate Sun position 
            sun_hrz = sun_dir.hrz(currentTime)
            (ra, dec) = sun_pos.apparent_equ(currentTime).format()
            strdict['sun_ra'] = "%-15s" % ra
            strdict['sun_dec'] = "%-15s" % dec
            strdict['sun_alt'] = "%6.2f" % sun_hrz.alt
            strdict['sun_az'] = "%6.2f" % sun_hrz.az
            try:
                if sun_hrz.alt > last_sun_hrz.alt:
                    sun_rising = True
                else:
                    sun_rising = False
            except NameError:
                pass
            last_sun_hrz = sun_hrz
            
            # calculate Jupiter posn
            jup_hrz = jup_dir.hrz(currentTime)
            (ra, dec) = jup_pos.apparent_equ(currentTime).format()
            strdict['jup_ra'] = "%-15s" % ra
            strdict['jup_dec'] = "%-15s" % dec
            strdict['jup_alt'] = "%6.2f" % jup_hrz.alt
            strdict['jup_az'] = "%6.2f" % jup_hrz.az
            try:
                if jup_hrz.alt > last_jup_hrz.alt:
                    jup_rising = True
                else:
                    jup_rising = False
            except NameError:
                pass
            last_jup_hrz = jup_hrz
            
            # Galactic Center
            gal_hrz = gal_dir.hrz(currentTime)
            (ra, dec) = gal_pos.apparent_equ(currentTime).format()
            strdict['gal_ra'] = "%-15s" % ra
            strdict['gal_dec'] = "%-15s" % dec
            strdict['gal_alt'] = "%6.2f" % gal_hrz.alt
            strdict['gal_az'] = "%6.2f" % gal_hrz.az
            try:
                if gal_hrz.alt > last_gal_hrz.alt:
                    gal_rising = True
                else:
                    gal_rising = False
            except NameError:
                pass
            last_gal_hrz = gal_hrz
            
            # calculate Tau A (Crab) position
            tau_hrz = tau_dir.hrz(currentTime)
            (ra, dec) = tau_pos.apparent_equ(currentTime).format()
            strdict['tau_ra'] = "%-15s" % ra
            strdict['tau_dec'] = "%-15s" % dec
            strdict['tau_alt'] = "%6.2f" % tau_hrz.alt
            strdict['tau_az'] = "%6.2f" % tau_hrz.az
            try:
                if tau_hrz.alt > last_tau_hrz.alt:
                    tau_rising = True
                else:
                    tau_rising = False
            except NameError:
                pass
            last_tau_hrz = tau_hrz
            
            # calculate Cassiopeia A position
            cas_hrz = cas_dir.hrz(currentTime)
            (ra, dec) = cas_pos.apparent_equ(currentTime).format()
            strdict['cas_ra'] = "%-15s" % ra
            strdict['cas_dec'] = "%-15s" % dec
            strdict['cas_alt'] = "%6.2f" % cas_hrz.alt
            strdict['cas_az'] = "%6.2f" % cas_hrz.az
            try:
                if cas_hrz.alt > last_cas_hrz.alt:
                    cas_rising = True
                else:
                    cas_rising = False
            except NameError:
                pass
            last_cas_hrz = cas_hrz
            
            # calculate Cygnus A position
            cyg_hrz = cyg_dir.hrz(currentTime)
            (ra, dec) = cyg_pos.apparent_equ(currentTime).format()
            strdict['cyg_ra'] = "%-15s" % ra
            strdict['cyg_dec'] = "%-15s" % dec
            strdict['cyg_alt'] = "%6.2f" % cyg_hrz.alt
            strdict['cyg_az'] = "%6.2f" % cyg_hrz.az
            try:
                if cyg_hrz.alt > last_cyg_hrz.alt:
                    cyg_rising = True
                else:
                    cyg_rising = False
            except NameError:
                pass
            last_cyg_hrz = cyg_hrz
            
            # Refresh screen and wait 1 second
            ## Background
            screen.addstr( 0, 1, "==========================================================================")
            screen.addstr( 1, 1, "UTC :                                   MJD(UTC):                         ")
            screen.addstr( 2, 1, "Site:                                   LST:                              ")
            screen.addstr( 3, 1, "Long:                                   Lat:                              ")
            screen.addstr( 4, 1, "==========================================================================")
            screen.addstr( 5, 1, "----------- Sun --------------          ------------ Jupiter -------------")
            screen.addstr( 6, 1, "RA(Cur)                                 RA(Cur)                           ")
            screen.addstr( 7, 1, "Dec(Cur)                                Dec(Cur)                          ")
            screen.addstr( 8, 1, "Alt                                     Alt                               ")
            screen.addstr( 9, 1, "Az                                      Az                                ")
            screen.addstr(10, 1, "---- Galactic Center ---------          ---------- Tau A (Crab) ----------")
            screen.addstr(11, 1, "RA(Cur)                                 RA(Cur)                           ")
            screen.addstr(12, 1, "Dec(Cur)                                Dec(Cur)                          ")
            screen.addstr(13, 1, "Alt                                     Alt                               ")
            screen.addstr(14, 1, "Az                                      Az                                ")
            screen.addstr(15, 1, "---------- Cas A -------------          ------------ Cyg A ---------------")
            screen.addstr(16, 1, "RA(Cur)                                 RA(Cur)                           ")
            screen.addstr(17, 1, "Dec(Cur)                                Dec(Cur)                          ")
            screen.addstr(18, 1, "Alt                                     Alt                               ")
            screen.addstr(19, 1, "Az                                      Az                                ")
            screen.addstr(20, 1, "==========================================================================")
            ## Date/time/site
            screen.addstr(1,  6, strdict['utc'], inf)
            screen.addstr(2,  6, strdict['loc'], inf)
            screen.addstr(3,  6, strdict['long'], inf)
            screen.addstr(1, 50, strdict['mjdutc'], inf)
            screen.addstr(2, 50, strdict['lst'], inf)
            screen.addstr(3, 50, strdict['lat'], inf)
            ## Sun
            screen.addstr( 6, 10, strdict['sun_ra'],  std)
            screen.addstr( 7,  9, strdict['sun_dec'], std)
            screen.addstr( 8,  9, strdict['sun_alt'], dwn if sun_hrz.alt < 0 else up)
            try:
                screen.addstr( 8, 16, "^" if sun_rising else "v", rs if sun_rising else st)
            except NameError:
                pass
            screen.addstr( 9,  9, strdict['sun_az'],  dwn if sun_hrz.alt < 0 else up)
            ## Jupiter
            screen.addstr( 6, 51, strdict['jup_ra'],  std)
            screen.addstr( 7, 50, strdict['jup_dec'], std)
            screen.addstr( 8, 50, strdict['jup_alt'], dwn if jup_hrz.alt < 0 else up)
            try:
                screen.addstr( 8, 57, "^" if jup_rising else "v", rs if jup_rising else st)
            except NameError:
                pass
            screen.addstr( 9, 50, strdict['jup_az'],  dwn if jup_hrz.alt < 0 else up)
            ## GC
            screen.addstr(11, 10, strdict['gal_ra'],  std)
            screen.addstr(12,  9, strdict['gal_dec'], std)
            screen.addstr(13,  9, strdict['gal_alt'], dwn if gal_hrz.alt < 0 else up)
            try:
                screen.addstr(13, 16, "^" if gal_rising else "v", rs if gal_rising else st)
            except NameError:
                pass
            screen.addstr(14,  9, strdict['gal_az'],  dwn if gal_hrz.alt < 0 else up)
            ## Tau A
            screen.addstr(11, 51, strdict['tau_ra'],  std)
            screen.addstr(12, 50, strdict['tau_dec'], std)
            screen.addstr(13, 50, strdict['tau_alt'], dwn if tau_hrz.alt < 0 else up)
            try:
                screen.addstr(13, 57, "^" if tau_rising else "v", rs if tau_rising else st)
            except NameError:
                pass
            screen.addstr(14, 50, strdict['tau_az'],  dwn if tau_hrz.alt < 0 else up)
            ## Cas A
            screen.addstr(16, 10, strdict['cas_ra'],  std)
            screen.addstr(17,  9, strdict['cas_dec'], std)
            screen.addstr(18,  9, strdict['cas_alt'], dwn if cas_hrz.alt < 0 else up)
            try:
                screen.addstr(18, 16, "^" if cas_rising else "v", rs if cas_rising else st)
            except NameError:
                pass
            screen.addstr(19,  9, strdict['cas_az'],  dwn if cas_hrz.alt < 0 else up)
            ## Cyg A
            screen.addstr(16, 51, strdict['cyg_ra'],  std)
            screen.addstr(17, 50, strdict['cyg_dec'], std)
            screen.addstr(18, 50, strdict['cyg_alt'], dwn if cyg_hrz.alt < 0 else up)
            try:
                screen.addstr(18, 57, "^" if cyg_rising else "v", rs if cyg_rising else st)
            except NameError:
                pass
            screen.addstr(19, 50, strdict['cyg_az'],  dwn if cyg_hrz.alt < 0 else up)
            screen.refresh()
            time.sleep(1)
                
            # Check for keypress and exit if Q
            c = screen.getch()
            if (c > 0):
                if chr(c) == 'q': 
                    break
                if chr(c) == 'Q': 
                    break
            
        restorescreen()
        
    except KeyboardInterrupt:
        restorescreen()
